import numpy as np
import re


def extract_spad_data(data, mode):
    """
    Extract data from specific detector elements from SPAD-fcs data

    Parameters
    ----------
    data : 2D numpy array
        [Nt x 25] array with fcs data as a function of time assuming a 
        5x5 array detector
        if 26 columns: last one is assumed to be the sum over all channels.
    mode : scalar or string
        Either
        - a number i between 0-25: extract data detector element i
        - a number i between -1 and -26: sum all elements except for i
        - "sum": sum over all detector elements and all time points
        - "sum3": sum central 9 detector elements
        - "sum3_5D": sum central 9 detector elements in [z, y, x, t, c] dataset
        - "sum5": sum all detector elements (= last column)
        - "outer": sum 16 detector elements at the edge of the array
        - "chess0": sum all even numbered detector elements
        - "chess1": sum all odd numbered detector elements
        - "upper_left": sum the 12 upper left detector elements
        - "lower_right": sum the 12 lower right detector elements
        - "sum5left": sum pixels 6, 10, 11, 12, 16
        - "sum5right", "sum5top", "sum5bottom"
        - "C3+7+8+125": C followed by a custom list of channel numbers
            of which the sum is returned.

    Returns
    -------
    array : numpy array
        Nx1 element vector with N the number of time points
        The vector contains the extracted and/or summed data.

    """
    
    # Check data type
    if isinstance(mode, str):
        if mode[0] == 'C':
            # list of channels to be summed, e.g. "C1+3+12+42"
            channels = [int(i) for i in re.findall(r'\d+', mode)]
            return np.sum(data[:, channels], axis=1)
        else:
            return switch_mode_string(data, mode)
    else:
        # number given
        if mode >= 0:
            return data[:, mode]
        else:
            return np.subtract(data[:, 25], data[:,np.mod(-mode, 25)])
   

def sum_all(data):
    # return total photon counts over all bins over all detector elements
    return np.sum(data[:, 25])
 
def central(data):
    # return central detector element data
    return data[:, 12]

def sum3(data):
    # sum detector elements 6, 7, 8, 11, 12, 13, 16, 17, 18
    datasum = np.sum(data[:, [6, 7, 8, 11, 12, 13, 16, 17, 18]], 1)
    return datasum

def sum5left(data):
    # sum detector elements 6, 10, 11, 12, 16
    datasum = np.sum(data[:, [6, 10, 11, 12, 16]], 1)
    return datasum

def sum5right(data):
    # sum detector elements 8, 12, 13, 14, 18
    datasum = np.sum(data[:, [8, 12, 13, 14, 18]], 1)
    return datasum

def sum5top(data):
    # sum detector elements 2, 6, 7, 8, 12
    datasum = np.sum(data[:, [2, 6, 7, 8, 12]], 1)
    return datasum

def sum5bottom(data):
    # sum detector elements 12, 16, 17, 18, 22
    datasum = np.sum(data[:, [12, 16, 17, 18, 22]], 1)
    return datasum

def sum3_5D(data):
    # sum over all time bins
    datasum = np.sum(data, 3)
    # sum detector elements 6, 7, 8, 11, 12, 13, 16, 17, 18
    datasum = np.sum(datasum[:, :, :, [6, 7, 8, 11, 12, 13, 16, 17, 18]], 3)
    return datasum

def allbuthot(data):
    # sum detector elements all but 1
    print("Not pixel 1")
    datasum = np.sum(data[:, [x for x in range(25) if x != 1]], 1)
    return datasum

def sum5(data):
    # sum all detector elements
    # datasum = data[:, 25]
    datasum = np.sum(data, 1)
    return datasum

def allbuthot_5D(data):
    # sum over all time bins
    datasum = np.sum(data, 3)
    # sum over all detector elements except hot
    datasum = np.sum(datasum[:, :, :, [0, 2, 3, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 16, 17, 18, 19, 20, 21, 22, 23, 24]], 3)
    return datasum

def outer(data):
    # smn 16 detector elements at the edge of the detector
    datasum = np.sum(data[:, [0, 1, 2, 3, 4, 5, 9, 10, 14, 15, 19, 20, 21, 22, 23, 24]], 1)
    return datasum
 
def chess0(data):
    # sum over all even numbered detector elements
    datasum = np.sum(data[:, 0:25:2], 1)
    return datasum

def chess1(data):
    # sum over all odd numbered detector elements
    datasum = np.sum(data[:, 1:25:2], 1)
    return datasum

def chess3a(data):
    # sum over all even numbered detector elements in the central 3x3 matrix
    datasum = np.sum(data[:, [6, 8, 12, 16, 18]], 1)
    return datasum

def chess3b(data):
    # sum over all odd numbered detector elements in the central 3x3 matrix
    datasum = np.sum(data[:, [7, 11, 13, 17]], 1)
    return datasum

def upper_left(data):
    # sum over detector elements 0, 1, 2, 3, 4, 5, 6, 7, 8, 10, 11, 15
    datasum = np.sum(data[:, [0, 1, 2, 3, 4, 5, 6, 7, 8, 10, 11, 15]], 1)
    return datasum

def lower_right(data):
    # sum over detector elements 9, 13, 14, 16, 17, 18, 19, 20, 21, 22, 23, 24
    datasum = np.sum(data[:, [9, 13, 14, 16, 17, 18, 19, 20, 21, 22, 23, 24]], 1)
    return datasum

def switch_mode_string(data, mode):
    switcher = {
        "central": central,
        "center": central,
        "sum3": sum3,
        "sum3_5D": sum3_5D,
        "allbuthot": allbuthot,
        "allbuthot_5D": allbuthot_5D,
        "sum5": sum5,
        "sum": sum_all,
        "outer": outer,
        "chess0": chess0,
        "chess1": chess1,
        "chess3a": chess3a,
        "chess3b": chess3b,
        "upper_left": upper_left,
        "lower_right": lower_right,
        "sum5left": sum5left,
        "sum5right": sum5right,
        "sum5top": sum5top,
        "sum5bottom": sum5bottom
    }
    # Get the function from switcher dictionary
    func = switcher.get(mode, "Invalid mode")
    # Execute the function
    return func(data)


# -*- coding: utf-8 -*-
"""
Created on Fri Apr 12 12:01:51 2019

@author: SPAD-fcs
"""

import pickle
from .checkfname import checkfname


def savevar(var, fname):
    fname = checkfname(fname, 'pickle')
    with open(fname, 'wb') as f:
        pickle.dump(var, f)


def openvar(fname):
    fname = checkfname(fname, 'pickle')
    with open(fname, 'rb') as f:
        data = pickle.load(f)
        return data



"""
Fitmodels fields:
    model                   str                 name of the model
    paramNames              list of str         names of the parameters
    paramDefvalues          list of numbers     default values for the parameters
    allparamDefvalues       list of numbers     default values for all parameters that the function requires
    fitfunctionParamUsed    list of numbers     indices of the parameters of the fit function that are actually used
    paramFactors10          list of numbers     powers of 10 with which the parameters have to be multiplied
    paramMinbound           list of numbers     minimum values for all parameters (must already include factors of 10)
    paramMaxbound           list of numbers     maximum values for all parameters (must already include factors of 10)
    fitfunctionName         function            name of the fit function
    
    Example: a fit function f(a, b, c, d) has 4 input parameters but parameter c must always be set to 3.14 in a given model
    Then:   paramNames  = ['a', 'b', 'd']
            paramDefvalues = [3, 9.2, -4] # default values for parameters a, b, d
            allparamDefvalues = [-1, -1, 3.14, -1] # set c to 3.14 and all other parameters to -1
            fitfunctionParamUsed = [0, 1, 3] # only parameters 0, 1, and 3 are used for the fit model
"""

class FitModel:
    def __init__(self):
        self.model = None
        self.shortlabel = None
        self.param_names = None
        self.param_fittable = None
        self.param_def_values = None
        self.all_param_def_values = None
        self.param_factors10 = None
        self.param_minbound = None
        self.param_maxbound = None
        self.fitfunction_name = None
        self.fitfunction_param_used = None
    
    def set_params(self, params):
        self.model                    = params.get("model")
        self.shortlabel               = params.get("shortlabel")
        self.param_names              = params.get("paramNames")
        self.param_fittable           = params.get("paramFittable")
        self.param_def_values         = params.get("paramDefvalues")
        self.all_param_def_values     = params.get("allparamDefvalues")
        self.param_factors10          = params.get("paramFactors10")
        self.param_minbound           = params.get("paramMinbound")
        self.param_maxbound           = params.get("paramMaxbound")
        self.fitfunction_name         = params.get("fitfunctionName")
        self.fitfunction_param_used   = params.get("fitfunctionParamUsed")
            
    
    def returnfitparam(self):
        return self.model, self.param_names, self.param_def_values, self.fitfunction_name, self.fitfunction_param_used, self.param_factors10, self.param_minbound, self.param_maxbound

    @property
    def num_param(self):
        if self.fitfunction_name is None and self.all_param_def_values is None:
            return 1
        if self.model == 'Free diffusion pair-correlation (global fit)':
            return 26
        return(len(self.param_minbound))



import pandas


def closefile(fname):
    """
    Close .h5 file

    Parameters
    ----------
    fname : string
        Path to and name of the file.

    Returns
    -------
    random string.

    """
    
    if fname[-3:] == ".h5":
        # h5 file
        fileToClose = pandas.HDFStore(fname)
        fileToClose.close()
        fileToClose = "random string"
    
    return(fileToClose)




import os

def list_files(directory='C:/Users/SPAD-fcs/OneDrive - Fondazione Istituto Italiano Tecnologia', filetype='bin', substr=False):
    """
    Find all files with file extension 'filetype' in the given directory.
    Subfolders included.

    Parameters
    ----------
    directory : path, optional
        Directory path string, e.g. 'C:\\Users\\SPAD-fcs'
    filetype : str, optional
        File extension. The default is 'bin'.
    substr : str or Boolean False, optional
        Only files that contain "substr" are returned
        Use False to not filter on substr. The default is False.

    Returns
    -------
    files : list
        List of file names.

    """
    
    # use current directory is empty string is given
    if directory == "":
        directory = os.getcwd()
    
    files = []
    # r=root, d=directories, f = files
    for r, d, f in os.walk(directory):
        for file in f:
            if '.' + filetype in file and (substr == False or substr in file):
                files.append(os.path.join(r, file))
        
    return files



import numpy as np
from .checkfname import checkfname
from tifffile import imwrite, imread
from matplotlib import cm

def array2tiff(data, fname, pxsize=1, dim="yxz", transpose3=True):
    """
    Write 2D or 3D array to tiff image file

    Parameters
    ----------
    data : np.array
        2D or 3D array with data (integer numbers int16)
        order: TZCYXS
        with    t   time
                c   channel.
    fname : string
        Name of the file to write to.
    pxsize : float, optional
        Pixel size [µm]. The default is 1.
    dim : TYPE, optional
        String with dimensions in image
        e.g. z stack of planar images: dim = "yxz"
        The order must be "tzcyxs". The same order must be used for data
        E.g. for a xy time series: dim="tyx" and 'data' is a 3D
        array with time, y, and x as 1st, 2nd, and 3r dimension
        The only exception is that for a 3D array also "yxz" is ok
        in combination with transpose3=True
        (which moves the 3rd dimension to the first to correct the
         order). The default is "yxz".
    transpose3 : boolean, optional
        Transpose array. The default is True.

    Returns
    -------
    Tiff image.

    """
    
    # check file extension
    fname = checkfname(fname, "tiff")
    
    # check number of images in data
    ndim = data.ndim
    if ndim >= 3 and transpose3:
        # transpose data to make 3rd dimension first
        data = np.transpose(data, (2, 0, 1))
    
        # order of dimensions is now TZCYXS
        dimAll = "tzcyxs"
        N = [1, 1, 1, 1, 1, 1]
        d = 0
        Ishape = np.shape(data)
        for i in range(6):
            if dimAll[i] in dim:
                N[i] = Ishape[d]
                d += 1
        
        data.shape = N  # dimensions in TZCYXS order
        
    data = data.astype('int16')
    imwrite(fname, data, imagej=True, resolution=(1./pxsize, 1./pxsize), metadata={'unit': 'um'})
    
    # add every image to same tiff file    
    #imsave(fname, data)
    
    print("Done.")
    

def array2rgbtiff(data, fname, cmap=cm.hot):
    """
    Write 2D array to tiff image file

    Parameters
    ----------
    data : np.array
        2D array with data (integer values).
    fname : string
        Name of the file to write to.
    cmap : matplotlib color map, optional
        color map. The default is cm.hot.

    Returns
    -------
    tiff image.

    Example:
        array2rgbtiff(data, "data_rgb.tiff", cmap=cm.viridis)
    """
    
    # check file extension
    fname = checkfname(fname, "tiff")
    
    # color map
    data = data.astype(int)
    dataMin = np.min(data)
    data -= dataMin
    dataMax = np.max(data)
    dataRange = dataMax + 1
    colors = cmap
    colorTable = colors(np.linspace(0, 1, int(dataRange)))
    colorTable = colorTable[:, 0:3]
    
    # image
    Ny = np.shape(data)[0]
    Nx = np.shape(data)[1]
    image = np.zeros((Ny, Nx, 3), dtype='uint8')
    for y in range(Ny):
        for x in range(Nx):
            image[y, x, :] = 255 * colorTable[data[y, x], :]
    
    #image = image.astype(uint8)
    imwrite(fname, image, photometric='rgb')
    
    print("Done.")
    
def tiff2array(fileList, Nx, Ny):
    """
    Read .tiff images and return 3D numpy array

    Parameters
    ----------
    fileList : list
        List of .tiff filenames to read.
    Nx : int
        Dimensions of the images.
    Ny : int
        Dimensions of the images.

    Returns
    -------
    imarray : np.array
        3D numpy array (Nf x Ny x Nx) with Nf the number of files.

    """
    
    Nf = len(fileList)
    imarray = np.zeros((Nf, Ny, Nx))
    for i in range(Nf):
        fname = fileList[i]
        im = imread(fname)
        imarray[i,:,:] = np.array(im)
    return imarray


import numpy as np

def file_to_analogcount(fname, dtype=">u8", print_info=False):
    """
    Read a bin file and returns an array with the decoded count for each measurement
    Args:
        fname: name of the file containing the data

    Returns:
        A numpy array of unsigned int16 os size N x 25 where N is the number of measurements 
    """
    try:
        raw = np.fromfile(fname, dtype=dtype)

    except:
        if print_info:
            print("Error reading binary file")
        return None

    elements = raw.shape[0]
    
    if print_info:
        print(f"Elements: {elements}")

    print(raw)

    out = np.zeros((elements , 2), dtype = 'int32')
    out[:,0] = np.bitwise_and(raw, int(2**32-1)) # 32 bits
    out[:,1] = np.bitwise_and(np.right_shift(raw, 32), int(2**32-1)) # 4 bits
    
    outplot = np.reshape(out, (1, 200, 200, 10, 2))
    
    outplot = np.squeeze(np.sum(outplot, 3))
    
    # plt.figure()
    # plt.imshow(outplot[:,:,1])
    # plt.colorbar()

    # plt.figure()
    # plt.plot(np.concatenate((outplot[100,:,0], outplot[101,:,0], outplot[102,:,0])))
    # plt.scatter(list(range(600)),np.concatenate((outplot[100,:,0], outplot[101,:,0], outplot[102,:,0])))
    # plt.plot(np.concatenate((outplot[100,:,1], outplot[101,:,1], outplot[102,:,1])))
    # plt.scatter(list(range(600)),np.concatenate((outplot[100,:,1], outplot[101,:,1], outplot[102,:,1])))


    return out, outplot


import numpy as np


class ATimesData:
    pass


def arrival_times_2_filtered_time_trace(data, filter_function, microbin_length, macrobin_length):
    """
    Convert a list of arrivalTimes to an intensity time trace I(t)

    Parameters
    ----------
    data : object
        Object with for each detector element field with a 2D int array, e.g.:
            data.det0 = np.array(N x 2)
                with N the number of photons
                First column: macro arrival times [a.u.]
                Second column: micro arrival times [a.u.]
            data.macrotime: macroTime [s]
            data.microtime: microTime [s].
    filter_function : np.array()
        np.array(M x Nf) with the filter functions
            M: number of lifetime bins
            Nf: number of filters, tyically 2 (1 fluor, 1 afterpulse)
                sum of each row = 1
        For multiple detector elements, and thus multiple filters,
        this variable is np.array(Ndet x M x Nf).
    microbin_length : float
        Duration of each bin of the lifetime histogram [s].
    macrobin_length : float
        Duration of each bin of the final intensity trace [s].

    Returns
    -------
    timeTraceF
        Object with for each filter function a np.array(K x L) with
            K the number of photon bins of duration binLength
            L the number of detector elements.

    """
    
    # get list of detector fields
    listOfFields = list(data.__dict__.keys())
    listOfFields = [i for i in listOfFields if i.startswith('det')]
    Ndet = len(listOfFields)
    
    # get total measurement time
    Mtime = 0
    for i in range(Ndet):
        dataSingleDet = getattr(data, listOfFields[i])
        Mtime = np.max((Mtime, dataSingleDet[-1, 0]))
    Mtime *= data.macrotime # [s]
    
    # number of time bins for binned and filtered intensity traces
    Nbins = int(np.ceil(Mtime / macrobin_length))
    
    # make sure filter_function is a 3D array
    if len(np.shape(filter_function)) == 2:
        filter_function = np.expand_dims(filter_function, 0)
    
    # create empty arrays to store output traces
    Nfilt = np.shape(filter_function)[-1]
    timeTraces = np.zeros((Ndet, Nbins, Nfilt), dtype=float)
    
    # go through each channel and create filtered intensity trace
    for det in range(Ndet):
        print("Calculating filtered intensity trace " + listOfFields[det])
        dataSingleDet = getattr(data, listOfFields[det])
        # calculate for each photon in which bin it should be
        photonMacroBins = np.int64(dataSingleDet[:,0] * data.macrotime / macrobin_length)
        photonMicroBins = np.int64(dataSingleDet[:,1] * data.microtime / microbin_length)
        for i in range(len(photonMacroBins)):
            for filt in range(Nfilt):
                timeTraces[det, photonMacroBins[i], filt] += filter_function[det, photonMicroBins[i], filt]

    return np.squeeze(timeTraces)


def atimes_filtered(data, filter_function, micro_bin=False):
    """
    Filter a list of arrivalTimes

    Parameters
    ----------
    data : object
        Object with for each detector element field with a 2D int array, e.g.:
            data.det0 = np.array(N x 2)
                with N the number of photons
                First column: macro arrival times [a.u.]
                Second column: micro arrival times [*]
            data.macrotime: macro time [s]
            data.microtime: micro time [s]
            data.microbintime: micro bin time [s].
    filter_function : np.array()
        np.array(M x Nf) with the filter functions
                M: number of lifetime bins
                Nf: number of filters, tyically 2 (1 fluor, 1 afterpulse)
                    sum of each row = 1
            For multiple detector elements, and thus multiple filters,
            this variable is np.array(Ndet x M x Nf).
    micro_bin : Boolean, optional
        True if micro arrival times [*] are in bin numbers
        False if micro arrival times [*] are in [a.u.]
            In this case, the bin numbers are calculated as
            bin = t * data.microtime / data.microbintime
            with data.microtime the microtime unit in s
            and data.microbintime the bin time in s. The default is False.

    Returns
    -------
    data : object
        Same object as input but data.det0 is now np.array(N x 2+Nf)
            For every detector element, Nf columns are added with 
            the filtered weights for the arrival times.

    """
    
    # get list of detector fields
    listOfFields = list(data.__dict__.keys())
    listOfFields = [i for i in listOfFields if i.startswith('det')]
    Ndet = len(listOfFields)
    
    # make sure filter_function is a 3D array (detector, microbin, filter function)
    if len(np.shape(filter_function)) == 2:
        filter_function = np.expand_dims(filter_function, 0)
    
    # number of filters
    Nf = np.shape(filter_function)[2]
    
    # number of time bins
    M = np.shape(filter_function)[1]
    
    # micro times normalization factor
    microN = 1
    if not micro_bin:
        microN = data.microtime / data.microbintime
    
    # go through each channel and create filtered intensity trace
    for det in range(Ndet):
        print("Calculating filtered photon streams " + listOfFields[det])
        # get photon streams single detector element
        dataSingleDet = getattr(data, listOfFields[det])
        # remove exessive columns which may already contain filtered photon streams
        dataSingleDet = dataSingleDet[:, 0:2]
        # calculate for each photon the filtered values
        photonMicroBins = np.int64(np.floor(dataSingleDet[:,1] * microN))
        photonMicroBins = np.clip(photonMicroBins, 0, M-1)
        for filt in range(Nf):
            filteredValues = np.expand_dims(np.take(np.squeeze(filter_function[det, :, filt]), photonMicroBins), 1)
            dataSingleDet = np.concatenate((dataSingleDet, filteredValues), axis=1)
        setattr(data, listOfFields[det], dataSingleDet)

    return data


#from functions.appearance import corr_long_label as cll
import numpy as np
import matplotlib.image as mpimg
import copy
from datetime import datetime

#current = os.path.dirname(os.path.realpath('analysis_settings.py'))
#parent = os.path.dirname(current)
#sys.path.append(parent)

from ..fcs.fcs2corr import fcs_av_chunks, fcs_crosscenter_av
from ..fcs.fcs_polar import g2polar, g2flow

"""
==============================================================================
FFSlib = [FFSexp1, FFSexp2, FFSexp3, etc.] with each element corresponding to 
one FFS experiment, i.e. 1 2d image and multiple FFS measurements
    
    FFSexp1 = [FFSfile1, FFSfile2, FFSfile3, etc.] with each element
    corresponding to one FFS measurement

FFSlib
     -> FFSexp: image
         -> FFSfile: single spot measurement
             -> metadata
             -> analysis
==============================================================================
"""

class FFSlib:
    """
    FFSlib is the main data object. It contains all images,
    each of which may contain multiple FFS experiments,
    and each experiment may contain multiple analyses.
    """
    def __init__(self,
                 lib = None,
                 active_image = None,
                 version = '1.1',
                 notes = '',
                 date_created = None,
                 date_modified = None):
        self.lib = lib if lib is not None else []
        self.active_image = active_image
        self.version = version
        self.notes = notes
        self.date_created = date_created or str(datetime.now())
        self.date_modified = date_modified or str(datetime.now())
    
    @property
    def num_images(self):
        """Returns the number of images in the library."""
        return len(self.lib)
    
    def add_image(self, image, fname, mdata=None):
        """Adds a new image object to the library."""
        if not fname:
            raise ValueError("Image must have a valid filename.")
        new_image_obj  = FFSimage()
        new_image_obj.image_name = fname
        new_image_obj.image = image
        if mdata is not None:
            new_image_obj.image_metadata = mdata
        new_image_obj.image_metadata.num_pixels = np.shape(image)[1]
        new_image_obj.image_metadata.num_lines = np.shape(image)[0]
        new_image_obj.imageRaw = image # may be deleted
        self.lib.append(new_image_obj)
    
    def add_random_image(self):
        self.add_image(np.zeros((5,5)), 'no_image_added')
    
    def get_image(self, image_nr='active'):
        # return the imageNr-th image object from the library
        # by default return current image object
        if isinstance(image_nr, str) and image_nr == 'active':
            image_nr = self.active_image
        if image_nr is None or self.num_images < 1:
            return None
        if image_nr > -1:
            image_nr = np.min((image_nr, self.num_images-1))
        image = self.lib[image_nr]
        return image
    
    def remove_image(self, image_nr='active'):
        if image_nr == 'active':
            image_nr = self.active_image
        if image_nr is not None and image_nr < self.num_images:
            del self.lib[image_nr]
            if self.num_images == 0:
                self.active_image = None
            else:
                self.active_image = 0
    
    def get_image_name(self, image_nr):
        image = self.get_image(image_nr)
        if image is None:
            return "No images added yet."
        return image.image_name
    
    def return_image(self, image_nr=-1):
        # return the image_nr-th image from the library
        image = self.get_image(image_nr)
        if image is None:
            return mpimg.imread('files/Cells_DEKegfp_75x75um.jpg')
        return image.image
    
    
class FFSimage:
    """
    FFSImage contains everything related to a single 2D image, 
    including metadata, image data, and associated FFS files.
    """
    def __init__(self):
        self.image_name  = "" # path to image
        self.image_metadata = FFSmetadata() # image metadata
        self.image = None
        self.image_raw = None
        self.ffs_list = [] # empty list that will be filled with FFSfile objects
        self.active_ffs = None
    
    @property
    def num_files(self):
        # return the number of FFS files for image
        return len(self.ffs_list)
    
    def change_image(self,  image, fname, mdata=None):
        self.image_name = fname
        self.image = image
        if mdata is not None:
            self.image_metadata = mdata
        self.image_metadata.num_pixels = np.shape(image)[1]
        self.image_metadata.num_lines = np.shape(image)[0]
        self.imageRaw = image # may be deleted
    
    def update(self, image_name=None, image=None, ffs_list=None):
        if image_name is not None:
            self.image_name = image_name
        if image is not None:
            self.image = image
        if ffs_list is not None:
            self.ffs_list = ffs_list
    
    def add_ffs_file(self, FFSfileObj):
        self.ffs_list.append(FFSfileObj)
    
    def remove_ffs_file(self, file_num):
        if file_num is not None and file_num < self.num_files:
            del self.ffs_list[file_num]
            
    def get_ffs_file(self, file_num='active'):
        if type(file_num) == str and file_num == 'active':
            file_num = self.active_ffs
        if file_num is not None and file_num < self.num_files:
            return self.ffs_list[file_num]
        return None
    
    def print_image_metadata(self):
        txt = ''
        for prop in ['num_pixels', 'num_lines', 'range_x', 'dwelltime', 'pxsize']:
            propvalue = getattr(self.image_metadata, prop)
            if propvalue is not None:
                if prop == 'num_pixels':
                    txt += str(int(propvalue))
                if prop == 'num_lines':
                    txt += ' x ' + str(int(propvalue)) + ' pixels\n'
                if prop == 'range_x':
                    txt += str(int(propvalue)) + ' x ' + str(int(propvalue)) + ' um\n'
                if prop == 'dwelltime':
                    txt += 'Dwelltime: ' + str(int(1e6*propvalue)) + ' us\n'
                if prop == 'pxsize':
                    txt += 'Pixel size: ' + "{:.3f}".format(propvalue) + ' um\n'
                
        return txt
        
        
class FFSfile:
    # each FFS file for each image contains a single FFSfile object
    def __init__(self,
                 fname="",  # path to file
                 label=None, # file nick name
                 metadata=None,
                 analysis_list=None, # list of correlations to calculate
                 timetrace=None,
                 airy=None,
                 active_analysis=None):
        self.fname = fname
        self.label = label 
        self.metadata = metadata if metadata is not None else FFSmetadata()
        self.analysis_list = analysis_list if analysis_list is not None else []
        self.timetrace = timetrace
        self.airy = airy
        self.active_analysis = active_analysis # choose which analysis is active by default
    
    @property
    def num_analyses(self):
        return len(self.analysis_list)
    
    @property
    def coords(self):
        return self.metadata.coords
    
    @property
    def duration(self):
        return self.metadata.duration
    
    @property
    def number_of_elements(self):
        if self.timetrace is None:
            return None
        if len(np.shape(self.timetrace)) > 1:
            # 2 dimensions
            return np.shape(self.timetrace)[1]
        return len(self.timetrace)
    
    def add_analysis(self, mode, resolution=10, chunksize=10, algorithm='multipletau', chunks_off="allon", active_analysis=-1):
        FFScorrObj = FFScorr()
        duration = self.metadata.duration
        if active_analysis == -1:
            self.active_analysis = self.num_analyses
        else:
            self.active_analysis = active_analysis
        if chunks_off == "allon":
            chunks_off = np.ones(np.clip(int(np.floor(duration / chunksize)), 1, None))
        FFScorrObj.analysis(mode, resolution, chunksize, chunks_off, algorithm)
        self.analysis_list.append(FFScorrObj)
    
    def remove_analysis(self, an_nr):
        if an_nr is not None and an_nr < self.num_analyses:
            del self.analysis_list[an_nr]
            if self.num_analyses == 0:
                self.active_analysis = None
            else:
                self.active_analysis = np.min((self.num_analyses-1, self.active_analysis))
    
    def copy_correlation(self, an_nr):
        # get analysis
        an_orig = self.get_analysis(analysis_num=an_nr)
        if an_orig is None:
            return
        # copy analysis and fit
        self.analysis_list.append(copy.deepcopy(an_orig))
        an = self.get_analysis(analysis_num=self.num_analyses-1)
        # remove all fits
        for i in range(an.num_fits):
            an.remove_fit(0)
    
    def use_fit_as_data(self, an_num, fit_num):
        # get analysis and fit object
        an_orig = self.get_analysis(analysis_num=an_num)
        if an_orig is None:
            return
        fits = copy.deepcopy(an_orig.return_fit_obj(fit_num=fit_num))
        if fits is None:
            return
        # copy analysis and fit
        self.analysis_list.append(copy.deepcopy(an_orig))
        an = self.get_analysis(analysis_num=self.num_analyses-1)
        # remove all fits
        for i in range(an.num_fits):
            an.remove_fit(0)
        # fill correlation analysis with fit as correlation data
        elements = an.settings.elements # central, sum3x3, sum5x5
        for element in elements:
            Gsingle = an.get_corr(element)
            fit = fits.fit_all_curves
            if Gsingle is not None and fit[0].fitfunction_label not in ['Model-free displacement analysis', 'Mean squared displacement']:
                for j in range(len(fit)):
                    if element == fit[j].data:
                        # fit found, add fitres to all chunks
                        for c in range(int(an.num_chunks(duration=self.metadata.duration))):
                            Gchunk = an.get_corr(element + '_chunk' + str(c))
                            fitres = fit[j].fitresult
                            fitrange = fit[j].fitrange
                            start = fitrange[0]
                            stop = fitrange[1]
                            xout = Gchunk[start:stop, 0]
                            yout = Gchunk[start:stop, 1] - fitres
                            Gout = np.zeros((len(xout), 2))
                            Gout[:,0] = xout
                            Gout[:,1] = yout
                            setattr(an.corrs, element + '_chunk' + str(c), Gout)
                            an.settings.update(chunks_off=an.settings.chunks_off, analysis=an)
                            
                        
    def get_analysis(self, analysis_num=-1):
        if analysis_num == -1:
            # return active analysis
            analysis_num = self.active_analysis
        if analysis_num is not None and analysis_num < self.num_analyses:
            return self.analysis_list[analysis_num]
        return None
    
    def update(self, fname=None, label=None, coords=None, timetrace=None, airy=None, active_analysis=None):
        if label is not None:
            self.label = label
        if coords is not None:
            self.metadata.coords = coords
        if timetrace is not None:
            self.timetrace = timetrace
        if airy is not None:
            self.airy= airy
        if active_analysis is not None:
            if active_analysis == 'None':
                active_analysis = None
            self.active_analysis = active_analysis


class FFSmetadata:
    def __init__(self,
                 num_pixels = None,
                 num_lines = None,
                 num_frames = None,
                 range_x = None,
                 range_y = None,
                 range_z = None,
                 num_datapoints = None,
                 hold_off_x5 = None,
                 hold_off = None,
                 time_resolution = None,
                 dwelltime = None,
                 duration = None,
                 pxsize = None,
                 coords = None
                 ):
        self.num_pixels = num_pixels
        self.num_lines = num_lines
        self.num_frames = num_frames
        self.range_x = range_x # µm
        self.range_y = range_y # µm
        self.range_z = range_z # µm
        self.num_datapoints = num_datapoints
        self.hold_off_x5 = hold_off_x5 # ns
        self.hold_off = hold_off # ns
        self.time_resolution = time_resolution # µs
        self.dwelltime = dwelltime # s
        self.duration = duration # s
        self.pxsize = pxsize # µm
        self.coords = coords # row, column number [y, x] of FFS position


class FFScorr:
    # for each file multiple FFScorr analysis objects can be added for different
    # types of analyses, e.g. spot-variation, iMSD, etc.
    def __init__(self,
                 mode=None,
                 settings=None,
                 corrs=None,
                 fits=None,
                 active_fit=None):
        self.mode = mode
        self.settings = settings if settings is not None else CorrSettings()
        self.corrs = corrs
        self.fits = fits if fits is not None else [] # list with fit results
        self.active_fit = active_fit # choose which fit to show by default
    
    @property
    def n_curves_mode(self):
        return len(self.settings.elements)
    
    @property
    def num_fits(self):
        return len(self.fits)
    
    def analysis(self, correlationObj, resolution, chunksize, chunks_off, algorithm, active_fit=-1, det_type='Genoa Instruments 5x5'):
        self.mode = correlationObj.mode
        
        self.settings.elements = correlationObj.elements
        self.settings.list_of_g = correlationObj.list_of_g
        self.settings.average = correlationObj.average
        
        self.settings.resolution = resolution
        self.settings.algorithm = algorithm
        self.settings.chunksize = chunksize # s
        self.settings.chunks_off = chunks_off
        if active_fit == -1:
            active_fit = self.num_fits
        self.active_fit = active_fit
    
    def num_chunks(self, duration):
        if duration is not None:
            N = int(np.floor(duration / self.settings.chunksize))
            return N
        return None
    
    def remove_fit(self, fit_num):
        if fit_num is not None and fit_num < self.num_fits:
            del self.fits[fit_num]
            if self.num_fits == 0 or self.active_fit is None:
                self.active_fit = None
            else:
                self.active_fit = np.min((self.num_fits-1, self.active_fit))
    
    def return_fit_obj(self, fit_num=-1):
        # by default return active fit object
        f = self.fits
        if fit_num == -1:
            fit_num = self.active_fit
        if fit_num is not None and len(f) > fit_num:
            return f[fit_num]
        return None
    
    def get_corr(self, corrtype="random"):
        # return average correlation central, sum3x3, sum5x5, etc.
        Gall = self.corrs
        if Gall is None:
            return None
        if corrtype == "random":
            excluded_keys = {"dwellTime", "crossCenterAv", "chunksOff", "chunks_off"}
            # Get the filtered list of keys
            keys = [key for key in list(Gall.__dict__.keys()) if key not in excluded_keys]
            Gsingle = getattr(Gall, keys[0])
        elif 'crossCenterAv' in corrtype:
            if 'chunk' in corrtype:
                chunk = corrtype[len('crossCenterAv'):]
                Gsingle = fcs_crosscenter_av(Gall, returnField=chunk, returnObj = False)
            else:
                Gsingle = fcs_crosscenter_av(Gall, returnObj = False)
        else:
            try:
                Gsingle = getattr(Gall, corrtype + "_averageX")
            except:
                try:
                    Gsingle = getattr(Gall, corrtype + "_average")
                except:
                    Gsingle = getattr(Gall, corrtype)
        return Gsingle
    
    def get_corr3D(self, N=9):
        # convert set of cross-correlations to 3D array [tau, ch1, ch2]
        try:
            Gsingle = self.get_corr('V0_H0')
        except:
            return None, None
        tau = Gsingle[:,0]
        G3d = np.zeros((len(tau), N, N))
        for i in range(N):
            for j in range(N):
                try:
                    G = self.get_corr('V' + str(j-N//2) + '_H' + str(i-N//2))
                    G3d[:, j, i] = G[:,1]
                except:
                    return None, None
        return G3d, tau
    
    def analysis_summary(self):
        algorithm = str(self.settings.algorithm) if self.settings.algorithm is not None else "unknown algorithm"
        return self.mode + " (settings: " + str(self.settings.resolution) + "/" + str(self.settings.chunksize) + "/" + algorithm + ")"
    
    def calc_corr(self):
        # calculate correlation only if not yet calculated before
        if self.corrs is None:
            return True
        return False
    
    def update(self, mode=None, settings=None, corrs=None, fits=None, active_fit=None):
        # mainly used to store calculated correlations in analysis object
        if mode is not None:
            self.mode = mode
        if settings is not None:
            self.settings = settings
        if corrs is not None:
            self.corrs = corrs
        if fits is not None:
            self.fits = fits
        if active_fit is not None:
            # use active_fit = 'None' to set active fit to None
            if active_fit == 'None':
                active_fit = None
            self.active_fit = active_fit
    
    def corr_param(self):
        # return all parameters needed to perform autocorrelation calculation
        return self.mode, self.settings.resolution, self.settings.chunksize
    
    def add_fit_analysis(self, fitfunctionmodel, fitarray, startvalues, fitrange=[1, -1]):
        # fitarray always has length 12 (11 parameters + weighted fit)
        mode = self.mode
        modelname = fitfunctionmodel.model
        if any(fitarray[:-1]) > 0 or modelname in ['Maximum entropy method free diffusion', 'Flow heat map', 'Asymmetry heat map', 'Model-free displacement analysis']:
            corrFitObj = CorrFit()
            corrFitObj.add_fit_analyses(mode, fitfunctionmodel, fitarray, startvalues, fitrange, self.settings.elements)
            self.fits.append(corrFitObj)
            self.active_fit = self.num_fits - 1
    
    def update_fit_analysis(self, fitfunctionmodel, fitarray, startvalues, fitrange=[1, -1], fit_num=-1):
        mode = self.mode
        if fit_num == -1:
            fit_num = self.active_fit
        if any(fitarray) > 0 or fitfunctionmodel.model in ['Maximum entropy method free diffusion', 'Asymmetry heat map', 'Flow heat map', 'Model-free displacement analysis']:
            corrFitObj = self.fits[fit_num]
            corrFitObj.fit_all_curves = [] # empty the object and fill it from scratch
            corrFitObj.add_fit_analyses(mode, fitfunctionmodel, fitarray, startvalues, fitrange, self.settings.elements)
                    
        
class CorrSettings():
    def __init__(self,
                 elements = None,
                 list_of_g = None,
                 resolution = None,
                 algorithm = None,
                 chunksize = None,
                 chunks_off = None,
                 average = None):
        self.elements   = elements # fields that are returned by fcs2corr
        self.list_of_g  = list_of_g # what is sent to fcs2corr
        self.resolution = resolution
        self.algorithm  = algorithm
        self.chunksize  = chunksize
        self.chunks_off = chunks_off
        self.average    = average # which cross-correlations should be averaged for flow analysis
    
    def update(self, elements=None, resolution=None, chunksize=None, chunks_off=None, analysis=None, algorithm=None, list_of_g=None):
        if elements is not None:
            self.elements = elements
        if list_of_g is not None:
            self.list_of_g = list_of_g
        if algorithm is not None:
            self.algorithm = algorithm
        if resolution is not None:
            self.resolution = resolution
        if chunksize is not None:
            self.chunksize = chunksize
        if chunks_off is not None:
            # if chunks_off is changed, also the average correlation is changed
            self.chunks_off = chunks_off
            G = analysis.corrs
            idx = np.nonzero(chunks_off)
            idx = list(idx[0]) # list of indices of good chunks
            try:
                G = fcs_av_chunks(G, idx)
                try:
                    G = fcs_crosscenter_av(G, returnField='_averageX')
                except:
                    pass
                analysis.update(corrs=G)
            except:
                pass
            

class CorrFit():
    def __init__(self):
        self.fit_all_curves = [] # list with each element consisting of 3 objects with fits for central, sum3, sum5
    
    @property
    def num_fitcurves(self):
        return len(self.fit_all_curves)

    def return_field(self, field):
        if field == "fitAllCurves":
            return self.fit_all_curves
        if field == "w0":
            w0 = []
            num_curves = len(self.fit_all_curves)
            for i in range(num_curves):
                w0.append(self.fit_all_curves[i].w0)
            return w0
        if field == "D":
            D = []
            num_curves = len(self.fit_all_curves)
            for i in range(num_curves):
                D.append(self.fit_all_curves[i].D)
            return D
        fits = self.fit_all_curves
        if len(fits) == 0:
            return None
        return fits[0].return_field(field)
    
    def return_all(self, field):
        fits = self.fit_all_curves
        if len(fits) == 0:
            return None
        data = []
        for i in range(len(fits)):
            data.append(getattr(fits[i], field))
        
        return data
    
    def fitrange(self):
        fits = self.fit_all_curves
        if len(fits) == 0:
            return None
        fit = fits[0]
        return fit.fitrange
    
    def fitresults(self, returntype="string"):
        # return 12 fit start values for the (3) fcs curves, either as strings or as 2d array
        num_curves = len(self.fit_all_curves)
        num_param = 12
        stv = ["NaN" for j in range(num_param)]
        bls = [False for j in range(num_param)]
        fitfunction = None
        fitres_array = np.zeros((num_param, num_curves))
        for i in range(num_curves):
            fit = self.fit_all_curves[i]
            fitfunction = fit.fitfunction_label
            power10 = fit.param_factors10
            for j in range(len(fit.paramidx)):
                fitabsv = fit.startvalues[fit.paramidx[j]] / power10[j]
                fitres_array[j, i] = fitabsv
                if np.abs(fitabsv) < 1e-2 or np.abs(fitabsv) > 999:
                    fitresString = str(fitabsv)[0:10]
                else:
                    fitresString = str(fitabsv)[0:4]
                if i == 0:
                    stv[j] = fitresString
                else:
                    stv[j] += fitresString
                if i < num_curves - 1:
                    stv[j] += ", "
                bls[j] = bool(fit.fitarray[fit.paramidx[j]])
        bls[-1] = bool(fit.fitarray[-1])
        if returntype == "string":
            return [stv, bls, fitfunction]
        else:
            return fitres_array
    
    def fitresults_mfda(self):
        # return scatter plot for model-free displacement analysis
        num_curves = len(self.fit_all_curves)
        N = 5 # pixels per dimension
        difftimesarray = np.zeros((N, N))
        corrvarray = np.ones((N, N))
        Nfound = 0
        for vert in range(N):
            for hor in range(N):
                # find the right fit
                for c in range(num_curves):
                    if self.fit_all_curves[c].data == 'V'+str(vert-int(np.floor(N/2)))+'_H'+str(hor-int(np.floor(N/2))):
                        fitr = self.fit_all_curves[c].fitresult
                        difftimesarray[vert, hor] = self.fit_all_curves[c].fitresult[0]
                        if len(fitr) > 1:
                            corrvarray[vert, hor] = self.fit_all_curves[c].fitresult[1]
                        Nfound += 1
        if Nfound > 0:
            return difftimesarray, corrvarray
        
        return None
    
    def fitresults_asymmetrymap(self):
        # return asymmetry heat map
        num_curves = len(self.fit_all_curves)
        fit = self.fit_all_curves[-1]
        
        if fit.fitresult is None:
            return
        
        N = int(len(fit.fitresult))
        
        if num_curves == 4:
            columnorder = ['Right', 'Up', 'Left', 'Down'] # square array detector
        elif num_curves == 6:
            columnorder = ['Right', 'UpRight', 'UpLeft', 'Left', 'DownLeft', 'DownRight'] # airy detector
        else:
            return
        
        allfits = np.zeros((N, len(columnorder)))
        columnNotFound = False
        for i in range(num_curves):
            try:
                c = columnorder.index(self.fit_all_curves[i].data)
                allfits[:, c] = self.fit_all_curves[i].fitresult
            except:
                columnNotFound = True
        
        if columnNotFound:
            return np.zeros((5,5)), columnNotFound
        
        z = g2polar(allfits)
    
        return z, columnNotFound
    
    def fitresults_flowmap(self):
        # return flow heat map
        num_curves = len(self.fit_all_curves)
        fit = self.fit_all_curves[-1] # color as a function of radius
        
        if fit.fitresult is None:
            return
        
        N = int(len(fit.fitresult))
        
        dets = ['square', 'airy']
        columnorder_square = ['Right', 'Up', 'Left', 'Down'] # square array detector
        columnorder_airy = ['UpRight', 'UpLeft', 'DownLeft', 'DownRight'] # airy detector
        
        for det, columnorder in enumerate([columnorder_square, columnorder_airy]):
            allfits = np.zeros((N, len(columnorder)))
            columnNotFound = False
            for i in range(num_curves):
                try:
                    c = columnorder.index(self.fit_all_curves[i].data)
                    allfits[:, c] = self.fit_all_curves[i].fitresult
                except:
                    columnNotFound = True
            if not columnNotFound:
                break
        
        if columnNotFound:
            return np.zeros((5,5)), [0,0], columnNotFound
        
        z, flow = g2flow(allfits, detector=dets[det])
        u = 2*flow[0]
        r = 2*flow[1]
    
        return z, [r, u], columnNotFound
    
    def fitresults_mem(self, tau, nparam=5):
        # return diffusion times distributions for MEM fit
        num_curves = len(self.fit_all_curves)
        fit0 = self.fit_all_curves[0]
        
        taumin = np.log10(tau[fit0.fitrange[0]])
        taumax = np.log10(tau[fit0.fitrange[1]-1])
        tauD = np.logspace(taumin, taumax, len(fit0.startvalues[0:-nparam]))
        
        fitresArray = np.zeros((len(tauD), num_curves))
        for i in range(num_curves):
            fit = self.fit_all_curves[i]
            fitresArray[:, i] = fit.startvalues[0:-nparam]
        
        fit = self.fit_all_curves[0]
        power10 = fit.param_factors10
        num_param = 12
        stv = ["NaN" for j in range(num_param)]
        bls = [False for j in range(num_param)]
        for j in range(len(fit.paramidx)):
            fitabsv = fit.startvalues[-1-nparam+fit.paramidx[j]] / power10[j]
            stv[j] = fitabsv
            bls[j] = bool(fit.fitarray[fit.paramidx[j]])
        stv[0] = len(tauD)
        
        return fitresArray, tauD, stv, bls
    
    def fitresults_msd(self):
        fit0 = self.fit_all_curves[0]
        D = fit0.startvalues[0]
        rho = 1e-3 * fit0.startvalues[3] # µm
        slope = 2 / rho**2 * D
        offset = fit0.startvalues[1]
        tauvar = fit0.fitresult
        tau = tauvar[0]
        var = tauvar[1]
        varfit = tau * slope + offset
        
        return tau, var, varfit
            
    def add_fit_analyses(self, mode, fitmodel, fitarray, startvalues, fitrange=[1, -1], data=[]):
        for i in range(len(data)):
            self.fit_all_curves.append(FitSingleObj(data[i], fitmodel, fitarray, startvalues[:,i], fitrange))
    

class FitSingleObj:
    def __init__(self, data, fitmodel, fitarray, startvalues, fitrange):
        # fitmodel is an object of the class fitModels
        
        Nparam = fitmodel.num_param + 1 # weighted fit
        self.minbound = np.array(fitmodel.param_minbound)
        self.maxbound = np.array(fitmodel.param_maxbound)
        self.param_factors10 = np.array(fitmodel.param_factors10)
        self.data = data # central, sum3, sum5, etc.
        self.fitfunction_label = fitmodel.model # more readable name of the fit function
        self.fitrange = fitrange
        self.fitresult = None # fit residuals
        
        paramind = fitmodel.fitfunction_param_used
        
        self.fitfunction = fitmodel.fitfunction_name # fit function used for the calculation (not a string but a function)
        fitarrayTemp = np.array([False for i in range(Nparam)])
        startvTemp = np.array(fitmodel.all_param_def_values).astype(float)
        if paramind is not None:
            for i in range(len(paramind)):
                fitarrayTemp[paramind[i]] = fitarray[i]
                startvTemp[paramind[i]] = float(startvalues[i]) * fitmodel.param_factors10[i]
            fitarrayTemp[Nparam-1] = fitarray[-1]
        else:
            fitarrayTemp = None
            startvTemp = None
        
        self.fitarray = fitarrayTemp
        self.startvalues = startvTemp
        self.paramidx = paramind # array indices with the parameters of interest (for startvalues and fitarray)
        self.w0 = None # beam waist (nm)
        self.D = None # diffusion coefficient (µm^2/s)
    
    def update(self, fitresult=None, startvalues=None, w0=None, D=None):
        # start values contains two lists [startv, fitv]
        # with startv containing the start values of both the fitted and unfitted parameters
        # fitv the 
        # start values are [M x 1] vector with M the 12 (or 7 for circFCS) fit parameters
        if fitresult is not None:
            self.fitresult = fitresult
        if startvalues is not None:
            self.startvalues = startvalues
        if w0 is not None:
            self.w0 = w0
        if D is not None:
            self.D = D


import pickle


def checkfname(fname):
    if len(fname) <= 7 or fname[-7:] != ".pickle":
        fname = fname + ".pickle"
    return fname


def savefig(h, fname):
    fname = checkfname(fname)
    pickle.dump(h, open(fname, 'wb'))


def openfig(fname):
    fname = checkfname(fname)
    pickle.load(open(fname, 'rb'))



import matplotlib.pyplot as plt
import numpy as np
import ntpath
from pathlib import Path
from os import  getcwd

from ..tools.plot_colors import plot_colors
from ..tools.csv2array import csv2array
from ..tools.find_nearest import find_nearest
from ..tools.list_files import list_files


class PyCorrfit_data():
    pass


def plot_pycorr_all(folderName=[]):
    if folderName == []:
        folderName = getcwd()
    folderName = folderName.replace("\\", "/")
    folderName = Path(folderName)
    fileList = list_files(folderName, 'csv')
    data = np.empty((0,3), float)
    # go through each file
    for file in fileList:
        if "chunk" not in file and "Gn" not in file and "sum3_average_fit_results" in file and "SPAD" in ntpath.basename(file):
            # file found
            print(ntpath.basename(file) + " found.")
            if "fit_results" in file:
                # file with fit found
                result = plot_pycorrfit(file, savefig=0)
                data = np.append(data, [[result.Gfitstart, result.tauD, result.chi2]], axis=0)
            else:
                # file with experimental G found
                plot_g_csv(file, savefig=0)
    return data
                
                
def plot_g_csv(file, info="", savefig="png"):
    """
    Plot .csv autocorrelation file

    Parameters
    ----------
    file : string
        .csv file with Nx2 matrix with tau, G.
    info : string, optional
        String with info about the measurement ('central', 'sum3', etc.)
        Needed for the color. The default is "".
    savefig : TYPE, optional
        0 to not save the figure, string with extension "png" or "eps"
        to save in this format. The default is "png".

    Returns
    -------
    data : np.array()
        Nx4 array with tau, G.

    """
    
    if "sum3" in file:
        info = "sum3"
    elif "sum5" in file:
        info = "sum5"
    elif "central" in file:
        info = "central"
    elif "ullr" in file:
        info = "ullr"
    elif "chessboard" in file:
        info = "chessboard"
        
    data = csv2array(file, ',')
    
    tau = data[:,0]
    G = data[:,1]
    taumin = np.min(tau[2:])
    
    plt.figure()
    plt.plot(tau, G, c=plot_colors(info), linewidth=1)
    plt.xscale('log')
    plt.xlim([taumin, 10])
    plt.ylim([np.min(G[2:]),np.max(G[2:])])
    plt.ylabel('G')
    plt.xlabel('tau (s)')
    
    plt.tight_layout()
    
    if savefig != 0:
        plt.savefig(file[0:-3] + savefig, format=savefig)
    
    return data


def plot_pycorrfit(file, info="", savefig=0, plot_tau=True):
    """
    Plot .asc file with PyCorrFit data

    Parameters
    ----------
    file : string
        .csv file with Nx4 matrix with tau, G, Gfit and residuals
        or object format:
            Gdata[:, 0] = tau
            Gdata[:, 1] = G
            Gdata[:, 2] = Gfit
            Gdata[:, 3] = Gres
            data.data = Gdata.
    info : string, optional
        String with info about the measurement ('central', 'sum3', etc.)
        Needed for the color. The default is "".
    savefig : int or string, optional
        0 to not save the figure
        file name with extension to save as "png" or "eps". The default is 0.
    plot_tau : boolean, optional
        Characteristic diffusion time, shown in the plot. The default is True.

    Returns
    -------
    allData : np.array
        Nx4 array with tau, G, Gfit and residuals.

    """
    
    if isinstance(file, str):
        # data given in file format
        if "sum3" in file:
            info = "sum3"
        elif "sum5" in file:
            info = "sum5"
        elif "central" in file:
            info = "central"
        elif "ullr" in file:
            info = "ullr"
        elif "chessboard" in file:
            info = "chessboard"
        allData = read_pycorrfit(file)
    else:
        # data given in object format
        allData = file
    
    tauD = allData.tauD # ms
    data = allData.data
    
    tau = data[:,0]
    G = data[:,1]
    Gfit = data[:,2]
    Gres = data[:,3]
    taumin = np.min(tau)
    taumax = np.min([0.5, np.max(tau)])
    print(taumax)
    
    # plot color
    if type(info) == str and info[0] == "C":
        pcolor = info
    else:
        pcolor = plot_colors(info)
    
    #plt.figure()
    f, (a0, a1) = plt.subplots(2, 1, gridspec_kw={'height_ratios': [3, 1]})
    a0.plot(tau, Gfit, c='k', linewidth=1)
    a0.scatter(tau, G, s=1, c=pcolor)
    a0.set_xscale('log')
    a0.set_xlim([taumin, taumax])
    a0.set_ylabel('G')
    a0.set_xticks([tl for tl in [1e-6, 1e-5, 1e-4, 1e-3, 1e-2, 1e-1] if tl > taumin and tl < taumax])
    
    # plot tau
    if plot_tau:
        ylim = a0.get_ylim()
        # if only single tau value is given, make tauD a list with 1 element
        if type(tauD) != list:
            tauD = [tauD]
        for i in range(len(tauD)):
            tauD2, idx = find_nearest(tau, 1e-3*tauD[i])
            GtauD2 = Gfit[idx]
            a0.plot([taumin, tauD2], [GtauD2, GtauD2], c='r', linewidth=0.7)
            a0.plot([tauD2, tauD2], [GtauD2, ylim[0]], c='r', linewidth=0.7)
        a0.plot([taumin, taumax], [0, 0], c='k', linewidth=0.7)
        a0.set_ylim(ylim)
    
    # plot residuals
    a1.plot(tau, Gres, c=pcolor, linewidth=1.0)
    a1.set_xlabel(r'$\tau$ (s)')
    a1.set_ylabel('residuals')
    a1.set_xscale('log')
    a1.set_xlim([taumin, taumax])
    a1.plot([taumin, taumax], [0, 0], c='k', linewidth=0.7)
    a1.set_xticks([tl for tl in [1e-5, 1e-4, 1e-3, 1e-2, 1e-1] if tl > taumin and tl < taumax])
    
    if savefig != 0:
        print("saving figure")
        if savefig[-3:] == 'svg':
            plt.rcParams['svg.fonttype'] = 'none'
        plt.tight_layout()
        plt.savefig(savefig, format=savefig[-3:])
    
    return allData


def read_pycorrfit(file):
    """
    Read header and data of .csv PyCorrFit output file

    Parameters
    ----------
    file : string
        String with path to .csv file.

    Returns
    -------
    outputdata : object
        Object with the tau, G, Gfit, Gres in data field, and separate
        fields for the fitted values n, SP, offset, and chi2.

    """
    
    # create object
    outputdata = PyCorrfit_data()
    # read .csv header
    f = open(file, "r")
    if f.mode == "r":
        contents = f.read()
        start = contents.find("Parameters:")
        [n, start] = read_pycorrfit_singleparam(contents, "#   n\t", "\n", start)
        [tauD, start] = read_pycorrfit_singleparam(contents, "_diff [ms]\t", "\n", start)
        [SP, start] = read_pycorrfit_singleparam(contents, "#   SP\t", "\n", start)
        [offset, start] = read_pycorrfit_singleparam(contents, "#   offset\t", "\n", start)
        start = contents.find("Fitting:", start)
        [chi2, start] = read_pycorrfit_singleparam(contents, "\t", "\n", start)
        [Gfitstart, start] = read_pycorrfit_singleparam(contents, "#   Ival start [ms]\t", "\n", start)
        [Gfitstop, start] = read_pycorrfit_singleparam(contents, "#   Ival end [ms]\t", "\n", start)
        outputdata.n = n
        outputdata.tauD = tauD
        outputdata.SP = SP
        outputdata.offset = offset
        outputdata.chi2 = chi2
        outputdata.Gfitstart = Gfitstart
        outputdata.Gfitstop = Gfitstop
    # load .csv file
    data = csv2array(file)
    # extract data
    tau = data[:,0]
    G = data[:,1]
    Gfit = data[:,2]
    Gres = G - Gfit
    outputdata.data = np.stack((tau, G, Gfit, Gres), axis=1)
    return outputdata


def read_pycorrfit_singleparam(contents, needle_start, needle_stop, startpos=0):
    """
    Read line in .csv file header

    Parameters
    ----------
    contents : string
        .csv file contents.
    needle_start : string
        First string to search for.
    needle_stop : string
        Second string to search for.
    startpos : int, optional
        Position to start searching. The default is 0.

    Returns
    -------
    n : int
        Number with the data in between needle_start and needle_stop.
    stop : int
        Stop position.

    """
    
    start = contents.find(needle_start, startpos)
    start += len(needle_start)
    if start == -1:
        print(needle_start + "not found.")
    stop = contents.find(needle_stop, start)
    n = np.float(contents[start:stop])
    return n, stop


def plot_pycorrfit3(fileC, Ac, file3, A3, file5, A5, savefig):
    """
    Plot .asc files with PyCorrFit results for SPAD data
    Data is normalized for visualization purposes

    Parameters
    ----------
    fileC : string
        .csv file with Nx4 matrix with tau, G, Gfit and residuals
        for the central pixel data.
    Ac : float
        Normalization constants (i.e. 1/N) for the central pixel
    file3 : string
        .csv file with Nx4 matrix with tau, G, Gfit and residuals
        for the sum 3x3 data.
    A3 : float
        Normalization constants (i.e. 1/N) for sum3
    file5 : string
        .csv file with Nx4 matrix with tau, G, Gfit and residuals
                    for the sum 5x5 data.
    A5 : float
        Normalization constants (i.e. 1/N) for sum5.
    savefig : boolean
        1 to save the figure as eps, 0 to not save the figure..

    Returns
    -------
    plots.

    """
    
    # Open and plot central pixel data file
    data = csv2array(fileC)
    taumin = np.min(data[:,0])
    tau = data[:,0]
    G = data[:,1] / Ac
    Gfit = data[:,2] / Ac
    plt.scatter(tau, G, s=1, c=plot_colors('central'))
    plt.plot(tau, Gfit, c=plot_colors('central'))
    plt.xscale('log')
    plt.xlim([taumin, 10])
    plt.ylabel('G (normalized)')
    plt.xlabel(r'$\tau$ (s)')
    
    # Open and plot sum 3 data file
    data = csv2array(file3)
    taumin = np.min(data[:,0])
    tau = data[:,0]
    G = data[:,1] / A3
    Gfit = data[:,2] / A3
    plt.scatter(tau, G, s=1, c=plot_colors('sum3'))
    plt.plot(tau, Gfit, c=plot_colors('sum3'))
    
    # Open and plot sum 5 data file
    data = csv2array(file5)
    taumin = np.min(data[:,0])
    tau = data[:,0]
    G = data[:,1] / A5
    Gfit = data[:,2] / A5
    plt.scatter(tau, G, s=1, c=plot_colors('sum5'))
    plt.plot(tau, Gfit, c=plot_colors('sum5'))
    
    font = {'size' : 12}
    plt.rc('font', **font)
    plt.legend(['central', 'sum3', 'sum5'])
    plt.tight_layout()
    
    if savefig:
        plt.savefig(fileC[0:-4] + '_and_sum3_sum5.eps', format='eps')
    


from .read_bh_spc import read_spc
from .arrivaltimes2timetrace import atimes2timetrace_bh
from .fcs2corr import fcs2corrsplit, plot_fcs_correlations
from .corr2csv import corr2csv

def bh_atimes2csv(fname, bintime=500, accuracy=50, split=20):
    """
    Convert Becker and Hickl fcs data to correlation curves stored in .csv files

    Parameters
    ----------
    fname : string
        File name with BH single photon arrival times (.asc file).
    bintime : float, optional
        Bin time [ns]. The default is 500.
    accuracy : int, optional
        Accuracy calculation autocorrelation. The default is 50.
    split : int, optional
        Number of fragments in which to split the time trace.
        For each fragment the autocorrelation is calculated, as
        well as the average correlation. The default is 20.

    Returns
    -------
    .csv file with all the correlations.

    """

    # load data
    data = read_spc(fname)

    # bin data
    dataBin = atimes2timetrace_bh(data, bintime)
    
    # calculate G
    G = fcs2corrsplit(dataBin, bintime/1000, ['det0'], accuracy, split)
    
    # plot G
    plot_fcs_correlations(G, ['det0_average'])
    
    # save as .csv file
    corr2csv(G, fname[0:-4])




import numpy as np
from .get_fcs_info import get_file_info
from .meas_to_count import file_to_fcs_count


class ATimesData:
    pass


def fcs2atimes(fname, macrotime, split=10):
    """
    Load SPAD-fcs data in chunks (of 10 s) and convert to arrivaltimes vectors

    Parameters
    ----------
    fname : string
        File name with the .bin data.
    macrotime : float
        Macrotime multiplication factor / dwell time [s].
    split : float, optional
        Number of seconds of each chunk to split the data into
        E.g. split=10 will divide a 60 second stream in 6 ten-second
        traces and calculate G for each individual trace. The default is 10.

    Returns
    -------
    data : object
        Object with following fields:
        data.det0chunk0    arrival times chunk 0 detector 0
        ...
        data.det24chunk10  arrival times chunk 10 detector 24
        data.macrotime      macrotime
        data.duration       measurement duration [s]

    """
    
    info = get_file_info(fname[:-4] + "_info.txt")
    dwellTime = info.dwellTime
    duration = info.duration
    
    data = ATimesData()
    data.macrotime = dwellTime
    data.duration = duration
    
    N = np.int(np.floor(duration / split)) # number of chunks

    chunkSize = int(np.floor(split / dwellTime))
    for chunk in range(N):
        # ---------------- CALCULATE CORRELATIONS SINGLE CHUNK ----------------
        print("+-----------------------")
        print("| Loading chunk " + str(chunk))
        print("+-----------------------")
        dataChunk = file_to_fcs_count(fname, np.uint8, chunkSize, chunk*chunkSize)
        if np.max(dataChunk) > 1:
            print("Max too high.")
        else:
            for det in range(np.shape(dataChunk)[1]):
                setattr(data, "det" + str(det) + "chunk" + str(chunk), np.nonzero(dataChunk[:,det])[0])
    
    return data


from pathlib import Path
from os import path as pth
from ..tools.h5data import h5data
from tifffile import TiffFile
import czifile

class infoObj:
    pass


def get_metafile_from_file(fname, metastring='_info.txt'):
    """
    Get metadata file from fcs data file
    If data file is stored in .bin file --> then metadata is in info.txt file
    If data file is stored in .h5 file --> then metadata is in same .h5 file
    If data file is stored in .tiff file --> then metadata is in same .tiff file

    Parameters
    ----------
    fname : string
        File name.
    metastring : string, optional
        String of the metafile. The default is '_info.txt'.

    Returns
    -------
    TYPE
        File name of the metadata.

    """
    
    head_tail = pth.split(fname)
    folder = head_tail[0]
    fname_stripped = head_tail[1]
    root_ext = pth.splitext(fname_stripped)
    fname_stripped = root_ext[0]
    fname_ext = root_ext[1]
    if fname_ext != '.bin':
        return fname
    metafile = pth.join(folder, fname_stripped + metastring)
    return metafile


def get_fcs_info(fname, parameter=['all']):
    """
    Get info from fcs info file (old version)

    Parameters
    ----------
    fname : TYPE
        File name [.txt].
    parameter : list of strings, optional
        Name of the parameter to return
        E.g. "READING SPEED [kHz]" will return the reading speed
        Leave blank to return an object containing all info. The default is ['all'].

    Returns
    -------
    info : object or number
        Either an object will all information
        Or the requested parameter.

    """
    
    # PARSE INPUT
    fname = fname.replace("\\", "/")
    fname = Path(fname)
    
    f = open(fname, "r", encoding="ISO8859-1") 
    if f.mode == "r":
        contents = f.read()  # in the case of error remove the encoding ="ISO8859-1"
        if parameter == ['all']:
            output = infoObj()
            info = float(get_fcs_info_singleparam(contents, "DURATION [s]"))
            output.duration = info
            info = float(get_fcs_info_singleparam(contents, "TOTAL NUMBER OF DATA POINTS"))
            output.numberOfDataPoints = info
            info = float(get_fcs_info_singleparam(contents, "HOLD-OFF [x5 ns]"))
            output.holdOffx5 = info
            output.holdOff = info * 5
            info = float(get_fcs_info_singleparam(contents, "READING SPEED [kHz]"))
            output.readingSpeed = info
            output.dwellTime = 1e-3 / info # s
            output.timeResolution = 1e6 * output.dwellTime # µs
            return output
        else:
            output = get_fcs_info_singleparam(contents, parameter)
            return float(output)


def get_file_info(fname, parameter=['all']):
    """
    Get info from fcs/imaging info file

    Parameters
    ----------
    fname : TYPE
        File name [.._info.txt] or [..h5].
    parameter : TYPE, optional
        Name of the parameter to return
        E.g. "READING SPEED [kHz]" will return the reading speed
        Leave blank to return an object containing all info
        For h5 files, it is always an object with all info. The default is ['all'].

    Returns
    -------
    finfo : object or number
        Either an object will all information
        Or the requested parameter.

    """
    
    head_tail = pth.split(fname)
    fname_stripped = head_tail[1]
    root_ext = pth.splitext(fname_stripped)
    fname_ext = root_ext[1]
    
    if fname_ext == '.txt':
        # read from text file
        finfo = get_file_info_txt(fname, parameter=['all'])
        return finfo
    elif fname_ext == '.h5':
        # read from h5 file
        finfo = get_file_info_h5(fname)
        return finfo
    elif fname_ext == '.tiff':
        # read from tiff file
        finfo = get_file_info_tiff(fname)
        return finfo
    elif fname_ext == '.czi':
        finfo = get_file_info_czi(fname)
        return finfo
    
def get_file_info_txt(fname, parameter=['all']):
    """
    Get fcs parameter info from metadata file

    Parameters
    ----------
    fname : string
        File name [.._info.txt].
    parameter : list of strings, optional
        Name of the parameter to return
        E.g. "READING SPEED [kHz]" will return the reading speed
        Leave blank to return an object containing all info. The default is ['all'].

    Returns
    -------
    info : object or number
        Either an object will all information
        Or the requested parameter.

    """
    
    # PARSE INPUT
    fname = fname.replace("\\", "/")
    fnameP = Path(fname)
    
    try:
        f = open(fnameP, "r", encoding="ISO8859-1") 
    except:
        f = None
            
    if f is not None and f.mode == "r":
        contents = f.read()
        
        if contents.find("READING SPEED") != -1:
            # info file is old, first version with reading speed
            output = get_fcs_info(fname)
            return output
        
        elif parameter == ['all']:
            output = infoObj()
            
            info = float(get_fcs_info_singleparam(contents, "NUMBER OF TIME BINS PER PIXEL"))
            output.numberOfTbinsPerPixel = info
            
            info = float(get_fcs_info_singleparam(contents, "NUMBER OF PIXELS"))
            output.numberOfPixels = info
            
            info = float(get_fcs_info_singleparam(contents, "NUMBER OF LINES"))
            output.numberOfLines = info
            
            info = float(get_fcs_info_singleparam(contents, "NUMBER OF FRAMES"))
            output.numberOfFrames = info
            
            info = float(get_fcs_info_singleparam(contents, "RANGE X [µm]"))
            output.rangeX = info
            
            info = float(get_fcs_info_singleparam(contents, "RANGE Y [µm]"))
            output.rangeY = info
            
            info = float(get_fcs_info_singleparam(contents, "RANGE Z [µm]"))
            output.rangeZ = info
            
            info = float(get_fcs_info_singleparam(contents, "TOTAL NUMBER OF DATA POINTS"))
            output.numberOfDataPoints = info
            
            info = float(get_fcs_info_singleparam(contents, "HOLD-OFF [x5 ns]"))
            output.holdOffx5 = info
            output.holdOff = info * 5
            
            info = float(get_fcs_info_singleparam(contents, "TIME RESOLUTION [µs]"))
            output.timeResolution = info
            
            output.dwellTime = output.timeResolution * output.numberOfTbinsPerPixel * 1e-6 # s
            output.duration = output.dwellTime * output.numberOfPixels * output.numberOfLines * output.numberOfFrames # s
            output.pxsize = output.rangeX / output.numberOfPixels
            
            return output
        else:
            output = get_fcs_info_singleparam(contents, parameter)
            return float(output)
    else:
        return f


def get_file_info_h5(fname):
    """
    Get metadata from fcs/imaging .h5 file
    
    Parameters
    ----------
    fname : string
        File name [..h5].

    Returns
    -------
    output : object or number
        An object will all information
        Or the requested parameter.

    """
    
    file = h5data(fname)
    
    output = infoObj()
    
    output.numberOfTbinsPerPixel = file.tbinpx()
    output.numberOfPixels = file.nx()
    output.numberOfLines = file.ny()
    output.numberOfFrames = file.nz()
    output.rangeX = file.rangex()        
    output.rangeY = file.rangey()        
    output.rangeZ = file.rangez()
    output.numberOfDataPoints = file.ndatapoints()
    output.timeResolution = file.tres()
    output.dwellTime = file.tres() * 1e-6 # s
    output.duration = file.duration() # s
    output.pxsize = file.pxsize()
    
    return output


def get_file_info_czi(fname):
    """
    Get metadata from czi file (under development)
    
    Parameters
    ----------
    fname : string
        File name [*.czi].

    Returns
    -------
    output : object or number
        An object will all information
        Or the requested parameter.

    """
    
    with czifile.CziFile(fname) as czi:
        # Access dimensions and shape
        dims_shape = czi.shape  # Shape of the image data
    
    output = infoObj()
    
    output.numberOfTbinsPerPixel = dims_shape[4] * dims_shape[7]
    output.numberOfPixels = 1
    output.numberOfLines = 1
    output.numberOfFrames = 1
    
    output.timeResolution = get_file_info_txt(fname[:-4] + '.txt', parameter="DWELL TIME [us]") # us
    if output.timeResolution is None:
        output.timeResolution = 1
    
    output.dwellTime = 1e-6 * get_file_info_txt(fname[:-4] + '.txt', parameter="DWELL TIME [us]") # s
    if output.dwellTime is None:
        output.timeResolution * 1e-6 # s
   
    output.duration = output.dwellTime * dims_shape[4]*dims_shape[7] # s
    
    return output

def get_file_info_tiff(fname):
    """
    Get metadata from fcs/imaging .tiff file
    
    Parameters
    ----------
    fname : string
        File name [*.tiff].

    Returns
    -------
    output : object
        An object will all information

    """
    output = infoObj()
    
    with TiffFile(fname) as tif:
        # Access the first image (page) and its tags
        tags = tif.pages[0].tags  # Retrieve metadata tags for the first image
        
        # check time resolution
        for tag in ["timeResolution", "time_resolution", "timeresolution", "dwellTime", "dwelltime", "dwell_time"]: # in us
            custom_tag = tags.get(tag)
            if custom_tag:
                output.timeResolution = custom_tag.value
                output.dwellTime = custom_tag.value * 1e-6 # s
            else:
                output.timeResolution = 1
                output.dwellTime = 1e-6 # s
            
        output.duration = max(tif.pages[0].shape) * output.dwellTime
    
    return output

def get_fcs_info_singleparam(contents, parameter):
    # get single parameter from fcs info file
    start = contents.find(parameter)
    start = start + len(parameter) + 1
    stop = contents.find("\n", start)
    if stop == -1:
        stop = len(contents)
    return contents[start:stop]



import numpy as np
import os
import argparse
import h5py
import sys
import tifffile
from ome_types import from_xml
from ..tools.list_files import list_files
from ..tools.array2tiff import array2tiff, array2rgbtiff
from .get_fcs_info import get_file_info, get_metafile_from_file, get_file_info_czi
import czifile
#import libttp

"""
This set of functions allows reading a binary file containing SPAD measurements
using only the file name. The parameters are extracted from the matrix using
the tags. The assumption is that the parameters are constant and that all the
frames are complete.
Author: Sebastian Acuna and Eli Slenders
"""

def file_to_count(fname, datatype=np.uint16, print_info=False, n_points=-1, n_offset=0):
    """
    Read a bin file and returns an array with the decoded count for each measurement
    Args:
        fname: name of the file containing the data

    Returns:
        A numpy array of unsigned int16 os size N x 25 where N is the number of measurements 
    """
    try:
        if n_points != -1:
            n_points = n_points * 2
            NbytesOffset = 16 * n_offset
            raw = np.fromfile(fname, dtype=">u8", count=n_points, offset=NbytesOffset)
        else:
            raw = np.fromfile(fname, dtype=">u8")

    except:
        if print_info:
            print("Error reading binary file")
        return None

    elements = raw.shape[0]
    positions = int(elements/2)
    raw_pos = np.reshape(raw, (positions, 2))
    if print_info:
        print(f"Elements: {elements}")
        print(f"Positions: {positions}")
        print(f"data table: {raw_pos.shape}")

    time_per_pixel_tag = np.bitwise_and(raw_pos[:,1], 0b1)
    idx = np.argmax(time_per_pixel_tag != time_per_pixel_tag[0]) # positions per time
    time_per_pixel = int(idx)
    if print_info:
        print(f"time per pixel: {time_per_pixel}")

    frame_tag = np.bitwise_and(np.right_shift(raw_pos[:,1], 2), 0b1)
    idx = np.argmax(frame_tag != frame_tag[0]) # positions per frame
    if idx == 0:
        if print_info:
            print("Unique frame")
        frames = 1
    else:
        frames = int(positions/idx) # TODO: check condition with larger dataset
        
    line_tag = np.bitwise_and(np.right_shift(raw_pos[:,1], 1), 0b1)
    idx = int(np.argmax(line_tag != line_tag[0])/time_per_pixel) # positions  per line
    if print_info:
        print(f"Positions per lines: {idx}")
    x = int(idx)
    y = int(positions/x/time_per_pixel/frames)

    if print_info:
        print(f"Dimensions: Y:{y}, X:{x}")

    out = np.zeros((positions , 25), dtype = datatype)

    matrix_to_count(raw_pos, out)

    return out, frames, y, x, time_per_pixel


def file_to_fcs_count(fname, datatype=np.uint16, n_points=-1, n_offset=0, h5dataset=None):
    """
    Read a *.bin, *.tiff, *.h5 file and returns an array with the decoded count
    Args:
        fname: name of the file containing the data
    Returns:
        A numpy array of unsigned int16 os size N x 25 where N is the number of measurements 
    """
    if fname[-3:] == '.h5':
        with h5py.File(fname, "r") as f:
            # Get the data
            if h5dataset is not None:
                data = f[h5dataset]
            else:
                key = None
                keys = list(f.keys())
                Nkeys = len(keys)
                if Nkeys > 1:
                    for i in range(Nkeys):
                        if 'data' in keys[i] and 'meta' not in keys[i]:
                            key = keys[i]
                if key is None:
                    key = keys[0]
                if 'data' in keys:
                    key = 'data'
                data = f[key]
                if len(np.shape(data)) > 2:
                    # flatten data to (t, c)
                    data = data[:]
                    data = data.reshape(-1, data.shape[-1])
                print(np.shape(data))
            if n_points == -1:
                out = data[n_offset:]
            else:
                out = data[n_offset:n_offset+n_points]
    elif fname.endswith('.tif') or fname.endswith('.tiff'):
        if n_points == -1:
            data = np.squeeze(tifffile.imread(fname))
            if len(data.shape) < 2:
                print("Image does not have 2 dimensions.")
                return None
            slices = [slice(n_offset, None)] + [slice(0, None)] + [0] * (len(data.shape) - 2)
            out = data[tuple(slices)]
        else:
            with tifffile.TiffFile(fname) as tif:
                ome_metadata_xml = tif.ome_metadata
                if ome_metadata_xml is None:
                    im = tifffile.imread(fname)
                    data_shape = np.shape(im)
                else:
                    ome_metadata = from_xml(ome_metadata_xml)
                    pixels = ome_metadata.images[0].pixels
                    px = pixels.size_x if pixels.size_x and pixels.size_x > 1 else None
                    py = pixels.size_y if pixels.size_y and pixels.size_y > 1 else None
                    pz = pixels.size_z if pixels.size_z and pixels.size_z > 1 else None
                    pc = pixels.size_c if pixels.size_c and pixels.size_c > 1 else None
                    pt = pixels.size_t if pixels.size_t and pixels.size_t > 1 else None
                    data_shape = (pt, pc, pz, py, px)
                    data_shape = tuple(x for x in data_shape if x is not None)
                if len(data_shape) < 2:
                    print("Image does not have 2 dimensions.")
                    return None
                slices = [slice(n_offset, n_offset + n_points)] + [slice(0, data_shape[1])] + [0] * (len(data_shape) - 2)
                out = tif.asarray()[tuple(slices)]
    
    elif fname.endswith('.czi'):
        # check if corresponding h5 exists, if not create it
        fname_h5 = fname[:-4] + '.h5'
        if not os.path.exists(fname_h5):
            print('converting czi file')
            _ = czi2h5(fname)
        with h5py.File(fname_h5, "r") as f:
            # Get the data
            data = f['data']
            print(np.shape(data))
            if n_points == -1:
                out = data[n_offset:]
            else:
                out = data[n_offset:n_offset+n_points]
            out = out[:]
    else:
        try:
            n_points = n_points * 2
            NbytesOffset = 16 * n_offset
            raw = np.fromfile(fname, dtype=">u8", count=n_points, offset=NbytesOffset)
    
        except:
            print("Error reading binary file")
            return None
    
        elements = raw.shape[0]
        print(f"Elements: {elements}")
        positions = int(elements/2)
        print(f"Positions: {positions}")
    
        print("Freeing memory")
        out = np.zeros((positions , 25), dtype = datatype)
        print("Done.")
        
        raw_pos = np.reshape(raw, (positions, 2))
        print(f"data table: {raw_pos.shape}")
    
        print("Converting data to counts")
        matrix_to_count(raw_pos, out)
        print("Done.")

    return out


def czi2h5(fname):
    """
    Convert a Zeiss czi file to h5. The h5 file is stored in the same location 

    Parameters
    ----------
    fname : str/path
        Path to the czi file

    Returns
    -------
    fname_h5 : string
        Path of the .h5 file.

    """
    out = np.squeeze(czifile.imread(fname))
    while len(np.shape(out)) > 3:
        out = out[:,0]
    out = np.transpose(out.reshape((32,out.shape[1]*out.shape[2])))
    fname_h5 = fname[:-4] + ".h5"
    with h5py.File(fname_h5, "w") as f:
        inf = f.create_group('configurationGUI_beforeStart')
        mdata = get_file_info_czi(fname)
        inf.attrs['nrep'] = 1
        inf.attrs['nframe'] = 1
        inf.attrs['ny'] = 1
        inf.attrs['nx'] = 1
        inf.attrs['range_z'] = 0
        inf.attrs['range_y'] = 0
        inf.attrs['range_x'] = 0
        inf.attrs['timebin_per_pixel'] = len(out)
        inf.attrs['time_resolution'] = mdata.timeResolution
        f.create_dataset('data',data=out, compression="gzip")
    return fname_h5

def bin2np(fname, datatype=np.uint16, Nchunks=-1, metadata=None, sum_tbins=True, read_chunks='all'):
    # r, y, x, t, c
    if Nchunks == -1:
        out, frames, y, x, time_per_pixel = file_to_count(fname, datatype)
        im = reshape_to_5d(out, frames, y, x, time_per_pixel)
    else:
        if metadata is None:
            metafile = get_metafile_from_file(fname)
            metadata = get_file_info(metafile)
        dwellTime = 1e-6 * metadata.timeResolution # s
        duration = metadata.duration
        split = duration / Nchunks
        chunkSize = int(np.round(split / dwellTime))
        if read_chunks == 'all':
            read_chunks = Nchunks
        for chunk in range(read_chunks):
            print('loading chunk ' + str(chunk) + '/' + str(Nchunks-1))
            [out, frames, y, x, time_per_pixel] = file_to_count(fname, datatype=datatype, n_points=chunkSize, n_offset=chunk*chunkSize)
            data = reshape_to_5d(out, frames, y, x, time_per_pixel)
            if sum_tbins:
                data = np.sum(data, 3) # sum over time bins --> f, y, x, c
            else:
                data = data[:,:,:,-1,:] # take only last time bin --> f, y, x, c (keeping the original structure requires too much memory)
            if chunk == 0:
                im = np.zeros((int(Nchunks*frames), y, x, 25))
            im[chunk*frames:(chunk+1)*frames, :, :, :] = data
            
    return im


def matrix_to_count(values, out):
    """
    Read an array of N measurements and write the count values in the out
    array

    Args:
        values: N x 2 unsigned int array with measurements
        out:    N x 25 unsigned int array for storing results

    Returns:
        The matrix out filled with the count
    """

    out[:,0] = np.bitwise_and(np.right_shift(values[:,0], 64 - 59), 0b1111) # 4 bits
    out[:,1] = np.bitwise_and(np.right_shift(values[:,0], 64 - 55), 0b1111) # 4 bits
    out[:,2] = np.bitwise_and(np.right_shift(values[:,0], 64 - 51), 0b1111) # 4 bits
    out[:,3] = np.bitwise_and(np.right_shift(values[:,0], 64 - 47), 0b1111) # 4 bits
    out[:,4] = np.bitwise_and(np.right_shift(values[:,0], 64 - 43), 0b1111) # 4 bits
    out[:,5] = np.bitwise_and(np.right_shift(values[:,0], 64 - 39), 0b1111) # 4 bits
    
    out[:,6] = np.bitwise_and(np.right_shift(values[:,1], 64 - 59), 0b11111) # 5 bits
    out[:,7] = np.bitwise_and(np.right_shift(values[:,1], 64 - 54), 0b111111) # 6 bits
    out[:,8] = np.bitwise_and(np.right_shift(values[:,1], 64 - 48), 0b11111) # 5 bits
    out[:,9] = np.bitwise_and(np.right_shift(values[:,1], 64 - 43), 0b1111) # 4 bits
    out[:,10] = np.bitwise_and(np.right_shift(values[:,1], 64 - 39), 0b1111) # 4 bits
    out[:,11] = np.bitwise_and(np.right_shift(values[:,1], 64 - 35), 0b111111) # 6 bits
    out[:,12] = np.bitwise_and(np.right_shift(values[:,1], 64 - 29), 0b1111111111) # 10 bits
    out[:,13] = np.bitwise_and(np.right_shift(values[:,1], 64 - 19), 0b111111) # 6 bits
    out[:,14] = np.bitwise_and(np.right_shift(values[:,1], 64 - 13), 0b1111) # 4 bits
    out[:,15] = np.bitwise_and(np.right_shift(values[:,1], 64 - 9), 0b1111) # 4 bits
    out[:,16] = np.right_shift(values[:,1], 64 - 5) # 5 bits
    
    out[:,17] = np.bitwise_and(np.right_shift(values[:,0], 64 - 35), 0b111111) # 6 bits 
    out[:,18] = np.bitwise_and(np.right_shift(values[:,0], 64 - 29), 0b11111) # 5 bits
    out[:,19] = np.bitwise_and(np.right_shift(values[:,0], 64 - 24), 0b1111) # 4 bits
    out[:,20] = np.bitwise_and(np.right_shift(values[:,0], 64 - 20), 0b1111) # 4 bits
    out[:,21] = np.bitwise_and(np.right_shift(values[:,0], 64 - 16), 0b1111) # 4 bits
    out[:,22] = np.bitwise_and(np.right_shift(values[:,0], 64 - 12), 0b1111) # 4 bits
    out[:,23] = np.bitwise_and(np.right_shift(values[:,0], 64 - 8), 0b1111) # 4 bits
    out[:,24] = np.bitwise_and(np.right_shift(values[:,0], 64 - 4), 0b1111) # 4 bits
    

def reshape_to_5d(count, frames, y, x, time_per_pixel):
    """
    Reshapes the 2D count matrix to a 5D array (frames, y, x, time, sensor)

    Args:
        count: N x 25 count matrix
        frames: number of frames contained in matrix
        y:
        x:
        time:

    Returns:
        A 5-D matrix with dimensions (frames, y, x, time, sensor)
    """

    return np.reshape(count, (frames, y, x, time_per_pixel, 25))

def reshape_to_6d(count, r, z, y, x, t=1, c=25):
    """
    Reshapes the data to a 6D array 

    Args:
        count: N x 25 count matrix
        frames: number of frames contained in matrix
        y:
        x:
        time:

    Returns:
        A 6-D matrix with dimensions (r, z, y, x, time, sensor)
    """

    return np.reshape(count, (r, z, y, x, t, 25))

def image2h5(fname, sum_time=True, save_time_ind=False):
    """
    Convert bin file to h5 file
        fname       file name
        sum_time    True to sum over all time bins, false otherwise
        save_time_ind Save all time frames in separate files
        TO DO:
            add metadata to file:
                data.pixelsize = 0.05
                data.pixelsizeU = 'um', etc.
    """
    print(fname)
    [out, frames, y, x, time_per_pixel] = file_to_count(fname)
    data = reshape_to_5d(out, frames, y, x, time_per_pixel)
    if np.ndim(data) == 4 and frames == 1 and sum_time:
        # 4D data set [y, x, time, ch] --> sum over time bins
        dataOut = np.sum(data, 2)
        dataOut = np.float64(dataOut)
        # channel must be first channel
        dataOut = np.transpose(dataOut, (2, 0, 1))
    elif np.ndim(data) == 5 and sum_time:
        # 5D data set [z, y, x, time, ch] --> sum over time bins
        dataOut = np.sum(data, 3)
        dataOut = np.float64(dataOut)
        dataOut = np.transpose(dataOut, (3, 0, 1, 2))
    else:
        print('not summed over time bins')
        # channel must be first channel
        dataOut = np.squeeze(data)
        if save_time_ind:
            dataOut = np.transpose(dataOut, (3, 0, 1, 2))
    
    dataOut = np.squeeze(dataOut)
    
    if type(save_time_ind) == bool and save_time_ind:
        for i in range(np.shape(dataOut)[-1]):
            print("Saving frame " + str(i))
            h5f = h5py.File(fname[:-4] + "_frame_" + str(i) + ".h5", 'w')
            h5f.create_dataset('dataset_1', data=dataOut[:,:,:,i])
            h5f.close()
    elif save_time_ind == "alternate":
        # channel, y, x, time
        Nt = np.shape(dataOut)[-1]
        for sumT in range(int(Nt/2)):
            print("Summing over " + str(sumT+1) + " frames")
            dataOutSum0 = np.squeeze(np.sum(dataOut[:,:,:,0:2*sumT+1:2], 3))
            dataOutSum1 = np.squeeze(np.sum(dataOut[:,:,:,1:2*sumT+2:2], 3))
            # store frame 0
            h5f = h5py.File(fname[:-4] + "_sum_" + str(sumT+1) + "_frame_0.h5", 'w')
            h5f.create_dataset('dataset_1', data=dataOutSum0)
            h5f.close()
            # store frame 1
            h5f = h5py.File(fname[:-4] + "_sum_" + str(sumT+1) + "_frame_1.h5", 'w')
            h5f.create_dataset('dataset_1', data=dataOutSum1)
            h5f.close()
    else:
        h5f = h5py.File(fname[:-3] + "h5", 'w')
        h5f.create_dataset('dataset_1', data=dataOut)
        h5f.close()
    
    return dataOut

def locseries2h5(fname, sumch = True, hot_pixels = []):
    """
    Convert bin file with time series of images to h5 file with channels summed
        fname       file name
    """
    print(fname)
    [out, frames, y, x, time_per_pixel] = file_to_count(fname)
    data = reshape_to_5d(out, frames, y, x, time_per_pixel)
    
    # t, x, y, c
    dataOut = np.squeeze(data)
    
    # sum over channels
    if sumch:
        Nch = np.shape(dataOut)[3]
        usedPixels = [i for i in range(Nch) if i not in hot_pixels]
        dataOut  = np.sum(dataOut[:,:,:,usedPixels], 3)
    
    dataOut = dataOut.astype(np.uint16)
    
    h5f = h5py.File(fname[:-3] + "h5", 'w')
    h5f.create_dataset('dataset_1', data=dataOut)
    h5f.close()
    
    return dataOut

    
def all_timetr_2_h5(folder=''):
    files = list_files(folder, filetype='ttr')
    filesh5 = list_files(folder, filetype='h5')
    for i, file in enumerate(filesh5):
        filesh5[i] = file[:-3] + '.ttr'
    for file in files:
        if file not in filesh5:
            print("converting " + file)
            #ttr2h5(file)
    print('Done')

def bin2h5(fname):
    """
    Convert bin file to h5 file with always 6D (r, z, x, y, t, c)
        fname       file name
            add metadata to file:
                data.pixelsize = 0.05
                data.pixelsizeU = 'um', etc.
    """

    [out, frames, y, x, time_per_pixel] = file_to_count(fname)
    data = reshape_to_6d(out, 1, frames, y, x, time_per_pixel, 25)
           
    # store data
    h5f = h5py.File(fname[:-4] + ".h5", 'w')
    h5f.create_dataset('dataset_1', data=data)
    h5f.close()
    print('Done')

def all_bin_images_2_tiff(folder):
    """
    Convert all bin files in a folder to tiff images
        folder      path to folder (use either \\ or / to go into a folder)
    """
    files = list_files(folder)
    for file in files:
        print("saving " + file)
        dummy = image2tiff(file)


def image2tiff(fname):
    """
    Convert bin file to tiff image file
        fname       file name
    """
    
    [out, frames, y, x, time_per_pixel] = file_to_count(fname)
    data = reshape_to_5d(out, frames, y, x, time_per_pixel)
    print(np.shape(data))
    info = get_file_info(fname[:-4] + '_info.txt')
    if np.ndim(data) == 4 and frames == 1:
        # 4D data set [y, x, time, ch] --> sum over time bins
        dataOut = np.sum(data, 2)
    elif np.ndim(data) == 5:
        # 5D data set [z, y, x, time, ch] --> sum over time bins and z
        dataOut = np.sum(data, 3)
        dataOut = np.sum(dataOut, 0)
    
    dataOut = np.float64(dataOut)    
    dataOut = np.squeeze(dataOut)
    
    print(np.shape(dataOut))
    
    array2tiff(dataOut, fname[:-4], pxsize=info.pxsize, dim="yxz", transpose3=True)
    array2rgbtiff(np.sum(dataOut, 2), fname[:-4] + '_RGB')


def all_bin_images_2_h5(folder, sum_time=True, save_time_ind=False):
    """
    Convert all bin files in a folder to h5 files
        folder      path to folder (use either \\ or / to go into a folder)
        sum_time    True to sum over all time bins, false otherwise
    """
    files = list_files(folder)
    for file in files:
        print("converting " + file)
        dummy = image2h5(file, sum_time, save_time_ind)


if __name__ == "__main__":

    parser = argparse.ArgumentParser(
        description="Converter from binary file to measurement matrix"
    )

    parser.add_argument(
        "binary",
        help="binary file name")

    args = parser.parse_args()
    fname = args.binary
    count, frames, y, x, time_per_pixel = file_to_count(fname)

    if count is None:
        print("Failed to process data. Closing.")
        sys.exit(0)

    file_name, extension = os.path.splitext(fname) # Get filename without extension

    print("Saving 5D matrix...", sep="")
    count5d = reshape_to_5d(count, frames, y, x, time_per_pixel)
    np.save(file_name + ".npy", count5d)
    print("Done.")

    







from .constants import constants


def calcconc(m, V, MM=368):
    """
    Calculate the molar concentration of a sample with mass m dissolved in
    volume V

    Parameters
    ----------
    m : float
        Mass of the solute [g].
    V : float
        Volume of the solution [L].
    MM : float, optional
        Molar mass of the solute [g/mole]
        Default set to 368 g/mole for Oregon Green. The default is 368.

    Returns
    -------
    c : float
        Concentration of the solution [mol/L].

    """
    
    Avogadro = constants('avogadro')
    
    # grams to moles of solute
    moles = m / MM
    
    # concentration [mole/L]
    c = moles / V
    
    # concentration [particles/L]
    c2 = c * Avogadro
    
    # concentration [particles/fL]
    c3 = c2 / 1e15
    
    # print results
    print('Concentration: {:.2e}'.format(c) + ' M = {:.2e}'.format(c2) + ' particles/L = {:.2e}'.format(c3) + ' particles/fL')
    
    return c



import numpy as np
import matplotlib.pyplot as plt


def phasor(H, normalize=True):
    """
    Caculate phasor values from given histogram
    g = 1 / N * sum(H * cos(f))
    s = 1 / N * sum(H * sin(f))

    Parameters
    ----------
    H : List or np.array
        List or np.array with histogram values.
    normalize : boolean, optional
        Normalize histogram. The default is True.

    Returns
    -------
    complex number describing the phasor.

    """
    
    Np = len(H)
    F = 2 * np.pi * np.linspace(0, Np-1, Np) / (Np - 1)
    
    norm = 1
    if normalize:
        norm = np.sum(H)
    
    g = np.sum(H * np.cos(F)) / norm
    s = np.sum(H * np.sin(F)) / norm
    
    z = complex(g, s)
    
    return(z)


def plotphasor(z):
    
    plt.figure()
    
    A = 1
    phi = np.arange(0, 2*np.pi, 2*np.pi/360)
    x1 = A * np.cos(phi)
    y1 = A * np.sin(phi)
    
    A = 0.5
    phi = np.arange(0, 2*np.pi, 2*np.pi/360)
    x2 = A * np.cos(phi) + 0.5
    y2 = A * np.sin(phi)
    
    plt.plot(x1, y1)
    plt.plot(x2, y2)
    plt.scatter(z.real, z.imag)
    plt.plot([-1, 1], [0, 0], color='k')
    plt.plot([0, 0], [-1, 1], color='k')
    
    plt.xlim([-1, 1])
    plt.ylim([-1, 1])
    plt.gca().set_aspect('equal', adjustable='box')    
    


import os
from .fcs2corr import Correlations
from ..tools.list_files import list_files
from ..tools.csv2array import csv2array

def load_g(fname_root, folder_name="", print_fnames=True):
    """
    Load correlations from .csv files into data object

    Parameters
    ----------
    fname_root : string
        'Root' of the file names. E.g. if the .csv files are named
        'FCS_beads_central', 'FCS_beads_sum3', and 'FCS_beads_sum5',
        then 'FCS_beads_' is the common root.
    folderName : string, optional
        Name of the folder to look into. The default is "".
    print_fnames : boolean, optional
        Print the names of the files that were found. The default is True.

    Returns
    -------
    G : correlations object
        Data object with G.central, G.sum3, G.sum5 etc. the correlation
        functions.

    """
    
    G = Correlations()
    files = list_files(folder_name, "csv", fname_root)
    for file in files:
        setattr(G, strip_g_fname(file, fname_root, print_fnames), csv2array(file, ','))
    G.dwellTime = 1e6 * csv2array(file, ',')[1, 0] # in µs
    print('--------------------------')
    print(str(len(files)) + ' files found.')
    print('--------------------------')
    return G


def strip_g_fname(fname, fname_root, print_fnames=True):
    fname = os.path.basename(fname)
    dummy, file_extension = os.path.splitext(fname)
    index = fname.find(fname_root)
    fname = fname[index + len(fname_root):-len(file_extension)]
    if print_fnames:
        print(fname)
    return fname
    


import numpy as np

def distance2detelements(det1, det2):
    """
    Calculate the distance between two SPAD detector elements in a.u.

    Parameters
    ----------
    det1 : int
        Number of the first detector element.
    det2 : int
        Number of the second detector element.

    Returns
    -------
    distance : float
        Distance between the two elements.

    """
    
    row1 = np.floor(det1 / 5)
    col1 = np.mod(det1, 5)
    
    row2 = np.floor(det2 / 5)
    col2 = np.mod(det2, 5)
    
    distance = np.sqrt((row2 - row1)**2 + (col2 - col1)**2)
    
    return distance


def spad_coord_from_det_numb(det):
    x = np.mod(det, 5)
    y = int(np.floor(det / 5))
    return([y, x])


def spad_shift_vector_crosscorr(vector=[0, 0], n=5):
    """
    Return a list of all cross-correlations between two channels that are
    displaced by a given shift vector. Used for STICS analysis.

    Parameters
    ----------
    vector : list, optional
        List with two elements [shifty, shiftx] in units of pixels.
        The default is [0, 0].
    n : int, optional
        Number of detector elements of the square detector. The default is 5
        for a 5x5 detector

    Returns
    -------
    CC : list of strings
        Each string correspond to a cross-correlation between two detector
        elements whose displacement is given by the input vector.

    """
    CC = []
    shifty = vector[0]
    shiftx = vector[1]
    for det1 in range(25):
        [det1y, det1x] = spad_coord_from_det_numb(det1)
        for det2 in range(25):
            [det2y, det2x] = spad_coord_from_det_numb(det2)
            # only use subset of pixels: central 3x3 or 5x5 region
            distanceFromCenter = np.array([np.abs(det2y-2), np.abs(det2x-2), np.abs(det1y-2), np.abs(det1x-2)])
            if det2y-det1y == shifty and det2x-det1x == shiftx and np.all(distanceFromCenter <= (n-1)/2):
                CC.append('det' + str(det1) + 'x' + str(det2))
    return CC
    



import numpy as np
from .fcs2corr import Correlations


def two_focus_fcs(tau, rhox, rhoy, rhoz, c, D, w0, w1, z0, z1):
    """
    Calculate the fcs cross correlation between two PSFs

    Parameters
    ----------
    tau : 1D numpy array
        Lag time [s].
    rhox : scalar
        Spatial shift in the x direction between two detectors [m].
        Corresponds to the spatial shifts in the sample space
    rhoy : scalar
        Spatial shift in the y direction between two detectors [m].
        Corresponds to the spatial shifts in the sample space.
    rhoz : scalar
        Spatial shift in the x direction between two detectors [m].
        Corresponds to the spatial shifts in the sample space. Usually 0.
    c : scalar
        Concentration of fluorophores/particles [/m^3].
    D : scalar
        Diffusion coefficient of the fluorophores/particles [µm^2/s].
    w0 : scalar
        Lateral 1/e^2 radius of the first PSF.
    w1 : scalar
        Lateral 1/e^2 radius of the second PSF.
    z0 : scalar
        Axial 1/e^2 value of the first PSF.
    z1 : scalar
        Axial 1/e^2 value of the second PSF.

    Returns
    -------
    G : correlations object
        Object with all autocorrelations.
    Garray : 2D numpy array
        Array with all autocorrelations..

    """

    factorW = 8 * D * tau + w0**2 + w1**2
    factorZ = 8 * D * tau + z0**2 + z1**2
    
    G0 = 2 * np.sqrt(2) / np.pi**(3/2) / c / factorW / np.sqrt(factorZ)
    
    expTerm = 16 * D * tau * (rhox**2 + rhoy**2)
    expTerm = expTerm + 2 * (z0**2 + z1**2) * (rhox**2 + rhoy**2)
    expTerm = expTerm + 2 * (w0**2 + w1**2) * rhoz**2
    expTerm = -1 * expTerm / factorW / factorZ
    expTerm = np.exp(expTerm)
    
    Gy = G0 * expTerm
    
    if type(Gy) == np.float64:
        Garray = np.zeros((1, 2))
    else:
        Garray = np.zeros((np.size(Gy, 0), 2))
    Garray[:, 0] = tau
    Garray[:, 1] = Gy

    G = Correlations()
    setattr(G, 'theory', Garray)

    return G, Garray


def simulate_two_focus_cross_center(tau, c, D, w, z, shift):
    G = Correlations()
    N = np.size(w, 0)
    for i in range(N):
        rhoy = np.floor(i / 5)
        rhox = np.mod(i, 5)
        [xx, Gtemp] = two_focus_fcs(tau, (rhox-2) * shift, (rhoy-2) * shift, 0, c, D, w[12], w[i], z[12], z[i])
        setattr(G, 'det12x' + str(i), Gtemp)
    return G


def simulate_spatialcorr(tau, rho, c, D, w0, z0):
    w0 = np.resize(w0, (5, 5))
    z0 = np.resize(z0, (5, 5))
    G = Correlations()
    if type(tau) == float:
        tau = [tau]
    Nt = len(tau)
    Gall = np.zeros((9, 9, Nt))
    for i in range(Nt):
        Gsinglet = np.zeros((9, 9))
        for shifty in np.arange(-4, 5):
            print('shifty: ' + str(shifty))
            for shiftx in np.arange(-4, 5):
                print(' shiftx: ' + str(shiftx))
                # go through all detector elements
                n = 0  # number of overlapping detector elements
                Gtemp = 0
                for detx in np.arange(np.max((0, shiftx)), np.min((5, 5+shiftx))):
                    print('  detx: ' + str(detx))
                    for dety in np.arange(np.max((0, shifty)), np.min((5, 5+shifty))):
                        print('   dety: ' + str(dety))
                        dummy, Gout = two_focus_fcs(tau[i], shiftx*rho, shifty*rho, 0, c, D, w0[dety, detx], w0[dety-shifty, detx-shiftx], z0[dety, detx], z0[dety-shifty, detx-shiftx])
                        Gtemp += Gout[0, 1]
                        n += 1
                Gtemp /= n
                Gsinglet[shifty+4, shiftx+4] = Gtemp
        Gall[:,:, i] = Gsinglet
    G.autoSpatial = Gall
    G.dwellTime = tau[0]
    return G




import numpy as np
from ..tools.savevar import savevar


class DataObj:
    pass

def read_asc(fname, store_pickle=False):
    """
    Read an ascii BH fcs file which contains the autocorrelation curve and
    return this curve

    Parameters
    ----------
    fname : string
        *asc file name.
    storePickle : boolean, optional
        Store data in pickle file. The default is False.

    Returns
    -------
    data
        Object containing information about the fcs file.

    """
    
    # OPEN FILE
    print("Opening .asc file.")
    with open(fname, mode='r') as file:
        rawdata = file.read()
    print("File opened.")
    
    # CREATE DATA OBJECT
    data = DataObj()

    # SEARCH fcs VALUE
    start = rawdata.find("FCS_value")
    if start == -1:
        # no fcs data found
        pass
    else:
        start = rawdata.find("\n", start) + 1
        stop =  rawdata.find("*END", start)
        Npoints = rawdata.count("\n", start, stop)
        G = np.zeros([Npoints, 2])
        Gn = np.zeros([Npoints, 2])
        for i in range(Npoints):
            stop = rawdata.find(" ", start)
            tp = float(rawdata[start:stop])
            start = stop + 1
            stop = rawdata.find("\n", start)
            Gp = float(rawdata[start:stop])
            G[i, 0] = tp
            G[i, 1] = Gp
            Gn[i, 0] = 1e-6 * tp
            Gn[i, 1] = Gp - 1
            start = stop + 1
        data.G = G
        data.Gn = Gn
        
    
    if store_pickle:
        print("Storing .pickle file")
        savevar(data, fname[0:-4] + "_BHdata")
        print(".pickle file stored")
    
    return data



import numpy as np
from scipy.optimize import least_squares

def fit_polynomial2d(z, ox, oy, weights=1):
    """
    Fit 2D polynomial function

    Parameters
    ----------
    z : TYPE
        Matrix with function values.
    ox : int
        polynomial order in x.
    oy : int
        polynomial order in y.
    weights : int or np.array, optional
        Matrix with fit weights
        1 for unweighted fit. The default is 1.

    Returns
    -------
    fitresult : TYPE
        Result of the nonlinear least squares fit.

    """
    
    Ncoeff = ox * oy + 2
    
    fitparamStart = np.zeros((Ncoeff))
    
    lowerBounds = fitparamStart - 1e6
    upperBounds = fitparamStart + 1e6
    
    fitparamStart += 1
    Ny = np.shape(z)[0]
    Nx = np.shape(z)[1]
    fitparamStart[0] = Ny / 2
    fitparamStart[1] = Nx / 2
    
    fitresult = least_squares(fitfun_polynomial2d, fitparamStart, args=(ox, oy, z, weights), bounds=(lowerBounds, upperBounds))
    
    return fitresult
    


def fitfun_polynomial2d(fitparamStart, ox, oy, z, weights=1):
    """
    Calculate residuals 2D polynomial function

    Parameters
    ----------
    fitparamStart : np.array()
        List with starting values for the fit parameters:
        order: [x0, y0, A, sigma, offset]
        with    x0, y0      position of the Gaussian center
                A           amplitude of the Gaussian
                sigma       Standard deviation of the Gaussian
                            w0 = 2 * sigma
                            with w0 1/e^2 radius
                offset      dc offset.
    ox : int
        polynomial order in x.
    oy : int
        polynomial order in y.
    z : 2D np.array()
        Matrix with function values.
    weights : int or np.array, optional
        Matrix with fit weights
        1 for unweighted fit. The default is 1.

    Returns
    -------
    res : vector
        Vector with residuals: f(param) - y.

    """
    
    Ny = np.shape(z)[0]
    Nx = np.shape(z)[1]
    
    res = np.reshape(weights * (polynomial2d(z, ox, oy, fitparamStart) - z), [Nx*Ny, 1])
    res = np.ravel(res)

    return res

def polynomial2d(z, ox, oy, coeffs):
    """
    Calculate 2D polynomial function

    Parameters
    ----------
    z : 2D np.array()
        Matrix with function values.
    ox : int
        polynomial order in x.
    oy : int
        polynomial order in y.
    coeffs : list
        list of coefficients.

    Returns
    -------
    out : np.array()
        matrix with 2D polynomial.

    """
    
    
    Ny = np.shape(z)[0]
    Nx = np.shape(z)[1]
    
    meshgrid = np.meshgrid(np.linspace(0, Nx-1, Nx), np.linspace(0, Ny-1, Ny))
    x = meshgrid[0]
    y = meshgrid[1]
    
    x0 = coeffs[0]
    y0 = coeffs[1]
    c = coeffs[2:].reshape((ox, oy))
    
    out = np.polynomial.polynomial.polyval2d(x-x0, y-y0, c)
    
    return out




import numpy as np
from ..tools.find_nearest import find_nearest

def g2difftime(g, smoothing=5):
    """
    Estimate diffusion time from fcs correlation curve without fitting
    
    Parameters
    ----------
    g : 2D np.array()
        Array with [tau, G] values.
    smoothing : int
        Number indicating moving average window for smoothing

    Returns
    -------
    t : float
        estimated diffusion time (same units as input)
    Gs : float
        correlation value at the diffusion time
    
    """
    # smooth function
    Gsmooth = np.convolve(g[1:,1], np.ones(smoothing)/smoothing, mode='valid')
    tausmooth = np.convolve(g[1:,0], np.ones(smoothing)/smoothing, mode='valid')
    
    # calculate derivative
    Gdiff = Gsmooth[1:] - Gsmooth[0:-1]
    Gdiff *= -1
    
    # normalized cumulative sum
    Gdcum = np.cumsum(Gdiff)
    Gdcum /= np.max(Gdcum)
    
    # now G is a function that starts at 0 and goes to 1. Diff time can be
    # estimated as tau such that G(tau) = 0.5
    [dummy, idx] = find_nearest(Gdcum, 0.5)
    
    return tausmooth[idx], Gsmooth[idx]


import h5py
import numpy as np
from datetime import datetime

class DataFileParam:
    def __init__(self):
        self.Creator = 'Name of the person/setup'
        self.DataSetDirectoryName = 'DataSet'
        self.Description = 'Measurement description, e.g. imaging test fluor beads'
        self.DocumentType = 'Image' # e.g. Image, fcs, SPT
        self.DichroicList = ['Chroma ZT405/488/561/640rpc']
        self.EmList = ['BP500-550']
        self.ExList = ['488']
        self.ExPowerList = ['10']
        self.ExPowerUnit = ['uW']
        self.ExFrequency = ['80']
        self.ExFrequencyUnit = ['MHz']
        self.Filename = 'File name'
        self.NameList = ['Channel name, e.g. cytoplasm']
        self.NumberOfDataSets = [1]
        self.NumericalAperture = 1.4
        self.Objective = '100x/1.4 Leica oil immersion'
        self.pxDwellTime = 10 # total time for one pixel (so over all time bins)
        self.pxDwellTimeUnit = 'us'
        self.pxSizeX = 1
        self.pxSizeY = 1
        self.pxSizeZ = 1
        self.SPADholdOff = 100 # hold-off time
        self.SPADholdOffUnit = 'ns'
        self.SPADmodel = 'BCD cooled'
        self.SPADtemperature = -15
        self.SPADtemperatureUnit = 'degr C'
        self.RecordingDate = [str(datetime.now())[0:23]] # list of recording times for each frame
        self.Unit = 'um'

def np2hdf5(filename, data, dataformat='Tzyxtc', param=DataFileParam()):
    """
    Convert data array to .hdf5 file
    ===========================================================================
    Input       Meaning
    ---------------------------------------------------------------------------
    filename    Name of the .hdf5 file
    data        N dimensional np.array with the raw data
                Default 6 dim (T, z, y, x, t, c)
                If fewer dimensions, or different order, use dataformat
                parameter to describe what's inside
    dataformat  Order of the dimensions, e.g. if data is 2D xy image,
                dataformat will be 'xy'
    param       Object with the metadata parameters
    ===========================================================================
    Output      Meaning
    ---------------------------------------------------------------------------
    .hdf5 file with always a 6 dimensional array for the data
    ===========================================================================
    """
    
    # check file name extension
    if len(filename) <= 4 or filename[-4:] != 'hdf5':
        filename = filename + ".hdf5"
    
    # not used at the moment
    if np.max(data) > 255:
        datatype = 'uint16'
    else:
        datatype = 'uint8'
    
    # make sure data has 6 dimensions
    dataShape = np.shape(data)
    Ndim = len(dataShape)
    for i in range(6-Ndim):
        data = np.expand_dims(data, Ndim+i)
    
    # check order of dimensions
    defOrder = 'Tzyxtc'
    order = []
    newdim = 0
    for i in range(6):
        dim = defOrder[i]
        if dim in dataformat:
            order.append(dataformat.find(dim))
        else:
            order.append(Ndim+newdim)
            newdim += 1
    data = np.transpose(data, order)
    
    # calculate finger print
    dataShape = np.shape(data)
    Nc = int(np.sqrt(dataShape[5]))
    fp = np.reshape(np.sum(np.sum(np.sum(np.sum(np.sum(data, 4), 3), 2), 1), 0), (Nc, Nc)) #fp = np.reshape(np.sum(data, axis=(4, 3, 2, 1, 0)), (Nc, Nc))
    
    # ---------------------------- Add data to file ---------------------------
    h5f = h5py.File(filename, 'w')
    h5f.create_dataset('RawData', data=data, compression="gzip")
    h5f.create_dataset('FingerPrint', data=fp, compression="gzip")
    
    # ------------------------------- Metadata --------------------------------
	# use same structure of mattia's metadata
    md = h5f.create_group("Metadata")
    fields = list(param.__dict__)
    for field in fields:
        md.attrs[field] = getattr(param, field)

    h5f.close()


def hdf5_read(filename, read='RawData', squeeze=False):
    """
    Read .hdf5 file with image/fcs/SPT data
    ===========================================================================
    Input       Meaning
    ---------------------------------------------------------------------------
    filename    Name of the .hdf5 file
    read        Object to read
                    'RawData': read raw data
                    'FingerPrint': read finger print
    squeeze     Remove singleton dimenions
    ===========================================================================
    Output      Meaning
    ---------------------------------------------------------------------------
    (6) dimensional numpy array or metadata
        hdf5_read('file.hdf5', read='RawData')
        hdf5_read('file.hdf5', read='pxDwellTime')
    ===========================================================================
    """
    with h5py.File(filename, "r") as f:
        if read in list(DataFileParam().__dict__):
            # return metadata
            return f['Metadata'].attrs[read]
        
        # Get the data
        data = f[read]
        
        if read == 'RawData':
            if squeeze:
                return np.squeeze(data[:,:,:,:,:,:])
            return data[:,:,:,:,:,:]
        
        elif read == 'FingerPrint':
            return data[:,:]
        
        return data[:]
    
    return None



import h5py

# how to use
# file = h5data('myfilename.h5')
# file.pxsize()
#   0.040
#
# to do: detector hold-off

class h5data():
    def __init__(self, fname):
        self.fname = fname
    
    def rangex(self):
        # range in um
        with h5py.File(self.fname, "r") as f:
            return(f['configurationGUI_beforeStart'].attrs['range_x'])
    
    def rangey(self):
        # range in um
        with h5py.File(self.fname, "r") as f:
            return(f['configurationGUI_beforeStart'].attrs['range_y'])
    
    def rangez(self):
        # range in um
        with h5py.File(self.fname, "r") as f:
            return(f['configurationGUI_beforeStart'].attrs['range_z'])
                
    def tbinpx(self):
        # number of time bins per pixel
        with h5py.File(self.fname, "r") as f:
            return(f['configurationGUI_beforeStart'].attrs['timebin_per_pixel'])
    
    def tres(self):
        # time resolution in us
        with h5py.File(self.fname, "r") as f:
            return(f['configurationGUI_beforeStart'].attrs['time_resolution'])
    
    def pxdwelltime(self):
        # pixel dwell time in us
        return(self.tres() * self.tbinpx())
    
    def frametime(self):
        # frame time in s
        return(self.pxdwelltime() * self.nx() * self.ny() / 1e6)
    
    def framerate(self):
        # frame rate in Hz
        return(1 / self.frametime())
    
    def nx(self):
        # number of pixels in x direction
        with h5py.File(self.fname, "r") as f:
            return(f['configurationGUI_beforeStart'].attrs['nx'])
            
    def ny(self):
        # number of pixels in y direction
        with h5py.File(self.fname, "r") as f:
            return(f['configurationGUI_beforeStart'].attrs['ny'])
    
    def nz(self):
        # number of frames
        with h5py.File(self.fname, "r") as f:
            return(f['configurationGUI_beforeStart'].attrs['nframe'])
    
    def nrep(self):
        # number of frames
        with h5py.File(self.fname, "r") as f:
            return(f['configurationGUI_beforeStart'].attrs['nrep'])
    
    def pxsize(self): # also y and z
        # pixel size in x direction
        return(self.rangex() / self.nx())
    
    def nmicroim(self):
        # total number of microimages read during the measurement
        return(self.nx() * self.ny() * self.nz() * self.nrep() * self.tbinpx())
    
    def ndatapoints(self):
        # total number of words transferred from low level to high level
        # 2 words per microimage
        return(2 * self.nmicroim())
    
    def duration(self):
        # total measurement duration in s
        return(self.nmicroim() * self.tres() * 1e-6)
    
    
def printmetadata(file):
    if not isinstance(file, h5data):
        file = h5data(file)
    for prop in ['tbinpx', 'nx', 'ny', 'nz', 'nrep',
                 'tres', 'rangex', 'rangey', 'rangez',
                 'pxdwelltime', 'frametime', 'framerate', 'pxsize',
                 'ndatapoints', 'duration']:
        print(prop, end = '')
        print(' ' * int(14 - len(prop)), end = '')
        print(str(getattr(file, prop)()))


import numpy as np

def bindata(data, binsize):
    """
    Sum elements in "data" in groups of size "binsize", creating a new vector
    with summed data

    Parameters
    ----------
    data : np.array
        Vector with data.
    binsize : int
        Number of data points per bin.

    Returns
    -------
    bdata : np.array
        Binned data.

    """
    
    cumdata = np.array(np.zeros([1, np.size(data, 1)]), dtype="uint32")
    cumdata = np.append(cumdata, np.uint32(np.cumsum(data, axis=0)), axis=0)
    bdata = cumdata[0::binsize, :]
    
    for i in range(np.size(data, 1)):
        bdata[0:-1, i] = np.ediff1d(bdata[:, i])
        
    bdata = bdata[0:-1]
    
    return bdata


def bindata_chunks(data, binsize, printChunkNr=False):
    bdata = np.empty((0, np.size(data, 1)), int)
    N = np.size(data, 0)
    chunksize = np.min([int(binsize * np.floor(10e6 / binsize)), N])
    binsize = np.min((binsize, chunksize))
    Nchunks = int(np.floor(N/chunksize))
    for i in range(Nchunks):
        if printChunkNr:
            print("Chunk " + str(i+1) + " of " + str(Nchunks))
        newbindata = bindata(data[i*chunksize:(i+1)*chunksize, :], binsize)
        np.size(newbindata)
        bdata = np.append(bdata, newbindata, axis=0)
    return bdata


import numpy as np


def fcs_analytical(tau, N, tauD, SF, offset, A=0, B=0, alpha=1):
    """
    Calculate the analytical fcs autocorrelation function assuming 3D Gaussian
    diffusion without triplet state

    Parameters
    ----------
    tau : 1D numpy array
        Lag time [s]
    N : scalar
        Number of particles on average in the focal volume [dimensionsless]
        N = w0^2 * z0 * c * pi^(3/2)
        with c the average particle concentration
    tauD : scalar
        Diffusion coefficient of the fluorophores/particles [µm^2/s]
    SF : scalar
        Shape factor of the PSF.
    offset : scalar
        DESCRIPTION.
    A : scalar, optional
        Afterpulsing characteristics. Power law assumed: G = A * tau^B (with B < 0).
        The default is 0.
    B : scalar, optional
        Afterpulsing characteristics. The default is 0.
    alpha : scalar, optional
        Anomalous diffusion parameter (alpha = 1 for free diffusion). The default is 1.

    Returns
    -------
    Gy : 1D numpy array
        Vector with the autocorrelation G(tau).

    """

    # standard autocorrelation function
    Gy = 1 / N / (1 + (tau/tauD)**alpha) # lateral correlation
    Gy /= np.sqrt(1 + tau**alpha / (SF**2 * tauD**alpha)) # axial correlation
    Gy += offset # offset
    # power law component to take into account afterpulsing (see e.g. Buchholz, Biophys J., 2018)
    Gy += A * tau**B
    
    if type(Gy) == np.float64:
        Garray = np.zeros((1, 2))
    else:
        Garray = np.zeros((np.size(Gy, 0), 2))
    Garray[:, 0] = tau
    Garray[:, 1] = Gy

    return Gy


def fcs_2c_analytical(tau, N, tauD1, tauD2, F, alpha=1, T=0, tautrip=1e-6, SF=5, offset=0, A=0, B=0):
    """
    Calculate the analytical fcs autocorrelation function assuming 3D Gaussian
    diffusion with triplet state, afterpulsing and 2 components

    Parameters
    ----------
    tau : 1D numpy array
        Lag time [s] (vector).
    N : scalar
        Number of particles on average in the focal volume [dimensionsless]
        N = w0^2 * z0 * c * pi^(3/2).
        with c the average particle concentration
    tauD1 : scalar
        Diffusion time species 1 [s].
    tauD2 : scalar
        Diffusion time species 2 [s].
    F : scalar
        Fraction of species 1.
    alpha : scalar, optional
        Relative molecular brightness q2/q1. The default is 1.
    T : scalar, optional
        Fraction in triplet. The default is 0.
    tautrip : scalar, optional
        Residence time in triplet state [s]. The default is 1e-6.
    SF : scalar, optional
        Shape factor of the PSF. The default is 5.
    offset : scalar, optional
        Offset. The default is 0.
    A : scalar, optional
        Afterpulsing characteristics. The default is 0.
        Power law assumed: G = A * tau^B (with B < 0)
    B : scalar, optional
        Afterpulsing characteristics. The default is 0.

    Returns
    -------
    Gy : 1D numpy array
        Vector with the autocorrelation G(tau).

    """

    # amplitude
    Gy = N * (F + alpha*(1-F))**2
    Gy = 1 / Gy
    
    # triplet
    Gy *= (1 + (T * np.exp(-tau / tautrip)) / (1 - T))
    
    # diffusion
    Gy *= F / (1 + tau/tauD1) / np.sqrt(1 + tau/SF**2/tauD1) + alpha**2 * (1-F) / (1 + tau/tauD2) / np.sqrt(1 + tau/SF**2/tauD2)

    # offset
    Gy += offset

    # afterpulsing (see e.g. Buchholz, Biophys J., 2018)
    Gy += A * tau**B

    return Gy


def fcs_analytical_2c_anomalous(tau, N, tauD1, tauD2, alpha1, alpha2, F, T, tau_triplet, SF, offset):
    # amplitude
    Gy = 1 / N
    
    # triplet fraction
    Gy *= (1 + T / (1 - T) * np.exp(-tau/tau_triplet))
    
    # two anomalous components
    Gcomp1 = F * (1 + (tau/tauD1)**alpha1)**(-1) * (1 + (tau/tauD1)**alpha1/SF**2)**(-1)
    Gcomp2 = (1 - F) * (1 + (tau/tauD2)**alpha2)**(-1) * (1 + (tau/tauD2)**alpha2/SF**2)**(-1)
    
    # total
    Gy *= (Gcomp1 + Gcomp2)
    Gy += offset
    
    return Gy


def fcs_dualfocus(tau, N, D, w, SF, rhox, rhoy, offset, vx=0, vy=0):
    """
    Calculate the analytical fcs crosscorrelation function for dual focus fcs
    assuming 3D Gaussian and diffusion without triplet state
    Equation from Scipioni, Nat. Comm., 2018 and consistent with own Maple
    calculations

    Parameters
    ----------
    tau : 1D numpy array
        Lag time [s] (vector).
    N : scalar
        Number of particles on average in the focal volume [dimensionsless]
        N = w0^2 * z0 * c * pi^(3/2)
        with c the average particle concentration
        and w0 the effective focal volume (w0^2 + w1^2) / 2.
    D : scalar
        Diffusion coefficient of the fluorophores/particles [m^2/s].
    w : scalar
        Radius of the effective PSF, i.e. sqrt((w0^2 + w1^2) / 2)
        with w0 and w1 the 1/e^2 radii of the two PSFs. [m]
    SF : scalar
        Shape factor of the PSF.
    rhox : scalar
        Distance between the two detector elements in the horizontal direction [m].
    rhoy : scalar
        Distance between the two detector elements in the vertical direction [m].
    offset : scalar
        DC component of G.
    vx : scalar, optional
        Velocity in x direction. The default is 0.
    vy : scalar, optional
        Velocity in y direction. The default is 0.

    Returns
    -------
    G : 1D numpy array
        Vector with the autocorrelation G(tau).

    """
    
    tauD = w**2 / 4 / D
    G = N * (1 + tau/tauD) * np.sqrt(1 + tau/(tauD*SF**2))
    G = 1 / G
    G = G * np.exp(-((rhox - vx*tau)**2 + (rhoy - vy*tau)**2) / w**2 / (1 + tau/tauD))
    G += offset
    
    return G


def fcs_2c_2d_analytical(tau, N, tauD1, tauD2, F, alpha=1, T=0, tautrip=1e-6, offset=0, A=0, B=0):
    # amplitude
    Gy = N * (F + alpha*(1-F))**2
    Gy = 1 / Gy
    
    # triplet
    Gy *= (1 + (T * np.exp(-tau / tautrip)) / (1 - T))
    
    # diffusion
    Gy *= F / (1 + tau/tauD1) + alpha**2 * (1-F) / (1 + tau/tauD2)

    # offset
    Gy += offset

    # afterpulsing (see e.g. Buchholz, Biophys J., 2018)
    Gy += A * tau**B

    return Gy

def nanosecond_fcs_analytical(tau, A, c_ab, tau_ab, c_conf, tau_conf, c_rot, tau_rot, c_trip, tau_trip, tauD, SP):
    # source: Galvanetto et al., Nature, 2023
    G = A
    G *= (1 - c_ab * np.exp(-tau / tau_ab)) # antibunching
    G *= (1 + c_conf * np.exp(-tau / tau_conf)) # conformational dynamics
    G *= (1 + c_rot * np.exp(-tau / tau_rot)) # rotational dynamics
    G *= (1 + c_trip * np.exp(-tau / tau_trip)) # triplet
    G /= ((1 + tau/tauD) * np.sqrt(1 + tau / SP**2 / tauD))
    
    return G

def uncoupled_reaction_diffusion(tau, A, tauD, SP, f_eq, k_off):
    """
    Uncoupled reaction and diffusion model
    Assumes that tauD << 1/k_on
    See Mazza et al., ch 12, Monitoring Dynamic Binding of Chromatin Proteins
    In Vivo by Fluorescence Correlation Spectroscopy
    and Temporal Image Correlation Spectroscopy
    """
    G = A
    G /= ((1 + tau/tauD) * np.sqrt(1 + tau / SP**2 / tauD))
    G += (1-f_eq)*np.exp(-k_off * tau)
    
    return G



import matplotlib.pyplot as plt
import numpy as np
import matplotlib.animation as animation
from ..tools.checkfname import checkfname


def fcsdata2video(data, fname='data.mp4', ftime=100, sum_all=False):
    """
    Convert SPAD-fcs data to 5x5 pixel video

    Parameters
    ----------
    data : 2D numpy array
        Data variable, i.e. output from binfile2data
        np.array(Nt x 25) with Nt number of time frames.
    fname : string, optional
        File name. The default is 'data.mp4'.
    ftime : scalar, optional
        Frame time [ms]. The default is 100.
    sum_all : boolean, optional
        False for normal 5x5 pixel video
        True for 1x1 pixel video with the sum of all pixels
            (single-element-detector-like). The default is False.

    Returns
    -------
    Saved video.

    """
    
    Nt = np.size(data, 0)
    
    if sum_all:
        # replace data with sum
        for i in range(Nt):
            data[i,:] = data[i,:] * 0 + np.sum(data[i,:])
    
    ims = []
    Imin = np.min(data)
    Imax = np.max(data)
    
    fig = plt.figure()
    FigSize = 10.5 # must be 10.5 to make the array size and video resolution match??
    
    fig.set_size_inches(FigSize, FigSize, forward=False)
    ax = plt.Axes(fig, [0., 0., 1., 1.])
    ax.set_axis_off()
    fig.add_axes(ax)
    for i in range(Nt):
        im = ax.imshow(np.reshape(data[i,0:25], (5, 5)), vmin=Imin, vmax=Imax)
        ims.append([im])
    
    ani = animation.ArtistAnimation(fig, ims, interval=ftime, blit=True)
    
    fname = checkfname(fname, 'mp4')
    
    ani.save(fname)


def partpos2video(pos, fname='video.mp4', limits=[2, 4, 2, 4, 5, 7], ftime=100, part_size=5, shadow_size=1, psf=[0.15, 0.15, 0.45]):
    """
    Convert SPAD-fcs particle positions to video

    Parameters
    ----------
    pos : 3D numpy array
        [Np x 3 x Nf] data array with
        Np: number of particles
        3:  x, y, z coordinates of the particle (float)
        Nf: number of frames.
    fname : string, optional
        File name. The default is 'video.mp4'.
    limits : list of 6 numbers, optional
        Plot axes limits [xmin, xmax, ymin, ymax, zmin, zmax]. The default is [2, 4, 2, 4, 5, 7].
    ftime : scalar, optional
        Frame time [ms]. The default is 100.
    part_size : scalar, optional
        Size of the dots for the particles. The default is 5.
    shadow_size : scalar, optional
        Size of the shadows of the particles, use 0 for no shadows. The default is 1.
    psf : list of 3 numbers, optional
        [wx, wy, z0] array with psf size. The default is [0.15, 0.15, 0.45].

    Returns
    -------
    Saved video.

    """
    
    Nf = np.shape(pos)[2]
    Np = np.shape(pos)[0]
    
    # plot limits
    xmin = limits[0]
    xmax = limits[1]
    ymin = limits[2]
    ymax = limits[3]
    zmin = limits[4]
    zmax = limits[5]
    
    pos = np.asarray(pos)
    # shadow 1
    posShadow1 = np.copy(pos)
    posShadow1[:, 2, :] = zmin
    # shadow 2
    posShadow2 = np.copy(pos)
    posShadow2[:, 1, :] = ymax
    pos = np.concatenate((pos, posShadow1), axis=0)
    pos = np.concatenate((pos, posShadow2), axis=0)
    
    # set default color of particles to green
    colorList = np.zeros((3*Np, 3, Nf))
    colorList[0:Np, :, :] = np.swapaxes(np.tile(np.array([0/256, 256/256, 0/256]), (Np, Nf, 1)), 1, 2)
    # change color of particles in laser to green
    counter = 0
    for f in range(Nf):
        for p in range(Np):
            if np.abs(pos[p, 0, f] - 3) < psf[0] and np.abs(pos[p, 1, f] - 3) < psf[1] and np.abs(pos[p, 2, f] - 6) < psf[2]:
                counter += 1
                colorList[p, :, f] = np.array([0/256, 256/256, 0/256])
    
    sizeList = np.zeros((3*Np, Nf))
    sizeList[0:Np, :] = np.tile(np.array([part_size]), (Np, Nf))
    if shadow_size > 0:
        sizeList[Np:3*Np, :] = np.tile(np.array([shadow_size]), (2*Np, Nf))
    else:
        sizeList[Np:3*Np, :] = np.tile(np.array([0]), (2*Np, Nf))
    
    # do not show particles outside of plot boundaries
    sizeList[pos[:,0,:] > xmax] = 0
    sizeList[pos[:,0,:] < xmin] = 0
    sizeList[pos[:,1,:] > ymax] = 0
    sizeList[pos[:,1,:] < ymin] = 0
    sizeList[pos[:,2,:] > zmax] = 0
    sizeList[pos[:,2,:] < zmin] = 0
    # if particles outside of boundaries, also remove shadows
    sizeList[np.concatenate((sizeList[0:Np,:] == 0, sizeList[0:Np,:] == 0, sizeList[0:Np,:] == 0), axis=0)] = 0
    
    fig = plt.figure()
    ax = plt.axes(projection = "3d")
    #ax.set_xticklabels([])
    #ax.set_yticklabels([])
    #ax.set_zticklabels([])
    
    ims = []
    for i in range(Nf):
        #[x, y, z] = draw_ellipsoid(psf[0], psf[1], psf[2], plotFig=False)
        #ax.plot_surface(x+2, y+2, z+6, rstride=4, cstride=4, color='b', alpha=0.2)
        im = ax.scatter3D(pos[:, 0, i], pos[:, 1, i], pos[:, 2, i], color=colorList[:,:,i], s=sizeList[:,i])
        ax.set_xlim([xmin, xmax])
        ax.set_ylim([ymin, ymax])
        ax.set_zlim([zmin, zmax])
        ims.append([im])
    
    ani = animation.ArtistAnimation(fig, ims, interval=ftime, blit=True)
    
    fname = checkfname(fname, 'mp4')
    
    ani.save(fname)


def fcstraces2video(data, fname='I_vs_t_video.mp4', ftime=100, dwellTime=1, dpi_user=250, xlabel="Time [s]", ylabel="PCR [Hz]", fsize=22, ylimits="auto"):
    """
    Convert SPAD-fcs data to line plot video with intensity traces over time

    WARNING: running this function from Jupyter Notebooks returns a lower
    quality video

    Parameters
    ----------
    data : 2D numpy array
        Data variable, i.e. output from binfile2data
        np.array(Nt x 25) with Nt number of time frames.
    fname : string, optional
        File name. The default is 'I_vs_t_video.mp4'.
    ftime : scalar, optional
        Frame time [ms]. The default is 100.
    dwellTime : scalar, optional
        Measurement dwell time [µs]. The default is 1.
    dpi_user : scalar, optional
        Video quality [dots per inch]. The default is 250.
    xlabel : string, optional
        x axis label. The default is "Time [s]".
    ylabel : string, optional
        y axis label. The default is "PCR [Hz]".
    fsize : scalar, optional
        text font size. The default is 22.
    ylimits : string or list of two numbers, optional
        y limits: [ymin ymax] or "auto". The default is "auto".

    Returns
    -------
    Saved video

    """
    
    if len(np.shape(data)) == 1:
        Nch = 1
    else:
        Nch = np.shape(data)[1]
    Nt = np.shape(data)[0]
    
    # bin time
    binTime = 1e-6 * dwellTime
    
    # time vector
    time = list(range(0, Nt))
    time = [i * binTime for i in time]
    
    # rescale intensity values to frequencies
    PCRscaled = data / binTime / 1000 # kHz
    ymax = np.max(PCRscaled)
    
    fig = plt.figure()
    plt.rcParams.update({'font.size': fsize})

    # axis limits
    if ylimits=="auto":
        ymin = 0
        ymax = 1.1 * ymax
    else:
        ymin = ylimits[0]
        ymax = ylimits[1]
    
    ax = plt.axes(xlim=(0, 2*time[-1] - time[-2]), ylim=(ymin, ymax))
    plt.xlabel(xlabel)
    plt.ylabel(ylabel)
    plt.tight_layout()
    
    lines = [plt.plot([], [])[0] for _ in range(Nch)] #lines to animate
    
    def init():
        #init lines
        for line in lines:
            line.set_data([], [])
    
        return lines #return everything that must be updated
    
    def animate(i):
        #animate lines
        for j,line in enumerate(lines):
            line.set_data(time[0:i], PCRscaled[0:i, j])
    
        return lines #return everything that must be updated
    
    ani = animation.FuncAnimation(fig, animate, init_func=init, interval=ftime, blit=True, frames=Nt)
    
    fname = checkfname(fname, 'mp4')
    
    ani.save(fname, dpi=dpi_user)
    


import numpy as np

def get_elsum(n, s):
    """
    Return list of all elements within ring s
    E.g.    s = 0 --> central element
            s = 1 --> sum3x3
            s = 2 --> sum5x5

    Parameters
    ----------
    N : int
        Number of rows/columns of the array, typically 5.
    s : int
        Outer ring number, typically between 0 and 2.

    Returns
    -------
    list
        list of indices.

    """
    
    out = []
    for r in range(s+1):
        out = np.append(out, get_el_ring_array(n, r))
    return [int(x) for x in out]


def get_el_ring_array(n, r):
    """
    Return list of all elements of a square array that are in ring r around
    the center. E.g. for a 5x5 array, there are three rings, ring 0-2:
    2 2 2 2 2
    2 1 1 1 2
    2 1 0 1 2
    2 1 1 1 2
    2 2 2 2 2
    E.g. Ring 1 contains elements [6,7,8,11,13,16,17,18]

    Parameters
    ----------
    n : int
        Number of rows/columns of the array, typically 5.
    r : int
        Ring number, typically between 0 and 2.

    Returns
    -------
    TYPE
        list of indices.

    """
    
    dist = np.zeros((n,n))
    for i in range(n):
        for j in range(n):
            dist[i, j] = np.max([np.abs(i-np.floor(n/2)), np.abs(j-np.floor(n/2))])
    
    dist = np.reshape(dist, (n*n,1))
    
    return np.where(dist == r)[0]


import numpy as np
from skimage import measure
from imutils import contours
import cv2
import imutils
import matplotlib.pyplot as plt

def exppsf(image, region=20, plots=False, blur=11, threshold=[20, 255]):
    """
    Measure experimental PSF based on an image of beads.
    Multiple beads are automatically found, overlapped with each other and
    summed for each channel separately. The resulting array is returned.

    Parameters
    ----------
    image : np.array
        2D array with (x, y) image data or
        3D array with (x, y, c) image data.
    region : int, optional
        Length of the square (in px) around each bead. The default is 20 pixels.
    plots : boolean, optional
        Plot results. The default is False.
    blur : int, optional
        Amount of blurring that should be added first. The default is 11.
    threshold : list, optional
        List of two values for the lower and upper threshold. The default is [20, 255].

    Returns
    -------
    PSFsum : np.array
        3D array with for each channel the summed PSF.
    PSFs : np.array
        4D array (region, region, Nc, Npart) with the image of each
        particle separately.

    """
    
    # check dimensions
    imageShape = np.shape(image)
    Ny = imageShape[0]
    Nx = imageShape[1]
    if len(np.shape(image)) > 2:
        Nchannels = np.shape(image)[2]
        # sum over all channels
        imageSum = np.sum(image, 2)
    else:
        Nchannels = 1
        imageSum = image
    
    if plots:
        plt.figure()
        plt.imshow(imageSum)
    
    particles = find_particles(imageSum, blur=blur, threshold=threshold)
    
    if plots:
        plt.figure()
        plt.scatter(particles[:,0], particles[:,1])
    
    # sum over all PSFs for each channel
    Nparticles = len(particles)
    PSFs = np.zeros((region, region, Nchannels, Nparticles))
    r = int(np.ceil(region / 2))
    # check for boundary effects
    NpartReal = 0
    for i in range(Nparticles):
        y = int(particles[i, 0])
        x = int(particles[i, 1])
        if y > r and y < Ny-r and x > r and x < Nx-r:
            if Nchannels > 1:
                PSFs[:,:,:,NpartReal] = image[y-r:y+r,x-r:x+r,:]
            else:
                PSFs[:,:,:,NpartReal] = np.expand_dims(image[y-r:y+r,x-r:x+r], axis=2)
            NpartReal += 1
    PSFs = PSFs[:,:,:,0:NpartReal]
    PSFsum = np.sum(PSFs, 3)
    
    return PSFsum, PSFs
        

def find_particles(im, blur=3, threshold=[0, 255], nPx=10, returnAll=False):
    """
    Find particles in image

    Parameters
    ----------
    im : np.array()
        2D array with (x, y) image data.
    blur : int, optional
        Int, width (in px) of the Gaussian blur applied first. The default is 3.
    threshold : list, optional
        [min, max] values for thresholding. The default is [0, 255].
    nPx : int, optional
        Number of pixels. The default is 10.
    returnAll : boolean, optional
        Return 4 variables with all results. The default is False.

    Returns
    -------
    particles
        2D array with particle coordinates.

    """
    
    # blur image
    blurred = cv2.GaussianBlur(im, (blur, blur), 0)
    
    # change data type
    blurred = blurred / np.max(blurred) * 255
    blurred = blurred.astype(np.uint8)
    
    # apply threshold
    thresh = cv2.threshold(blurred, threshold[0], threshold[1], cv2.THRESH_BINARY)[1]
    thresh = cv2.erode(thresh, None, iterations=1)
    thresh = cv2.dilate(thresh, None, iterations=1)
    
    # find peaks
    labels = measure.label(thresh)
    mask = np.zeros(thresh.shape, dtype="uint8")
    # loop over the unique components
    for label in np.unique(labels):
    	# if this is the background label, ignore it
    	if label == 0:
    		continue
    	# otherwise, construct the label mask and count the
    	# number of pixels 
    	labelMask = np.zeros(thresh.shape, dtype="uint8")
    	labelMask[labels == label] = 255
    	numPixels = cv2.countNonZero(labelMask)
    	# if the number of pixels in the component is sufficiently
    	# large, then add it to our mask of "large blobs"
    	if numPixels > nPx:
    		mask = cv2.add(mask, labelMask)
    
    cnts = cv2.findContours(mask.copy(), cv2.RETR_EXTERNAL,	cv2.CHAIN_APPROX_SIMPLE)
    Ncnts = len(cnts[0])
    print(str(Ncnts) + ' particles found')
    if Ncnts > 0:
        cnts = imutils.grab_contours(cnts)
        cnts = contours.sort_contours(cnts)[0]
        particles = np.zeros((len(cnts), 2))
        # loop over the contours
        for (i, c) in enumerate(cnts):
        	# draw the bright spot on the image
            (x, y, w, h) = cv2.boundingRect(c)
            [(cX, cY), radius] = cv2.minEnclosingCircle(c)
            particles[i, 0] = int(cY)
            particles[i, 1] = int(cX)
    else:
        particles = None
        
    if returnAll:
        return particles, Ncnts, blurred, thresh
    
    return particles


from .constants import constants
import numpy as np


def stokes_einstein(D, T=293, visc=1e-3):
    """
    Calculate the diameter of particles based on the Stokes-Einstein equation

    Parameters
    ----------
    D : float
        Diffusion coefficient of the particles [in m^2/s].
    T : float, optional
        Temperature of the suspension [in K]. The default is 293.
    visc : float, optional
        Viscosity of the solvent [in Pa.s]. The default is 1e-3.

    Returns
    -------
    d : float
        Diameter of the particles [in m].

    """
    
    kb = constants('boltzmann')
    r = kb * T / 6 / np.pi / visc / D
    d = 2 * r
    return d



def read_binfile(fname, start_index=0, length=1024):
    """
    Read a binary fcs data file, containing a list of U64 numbers that
    represent the photon counts per time bin of 1 µs. Each pair of two numbers
    contains the information of all 25 detector elements in a single time bin.

    Parameters
    ----------
    fname : string
        Binary file name.
    startIndex : int, optional
        Position in the file where the reading is started. The default is 0.
    length : int, optional
        Number of bytes to read (Note: 8 bytes/U64). The default is 1024.

    Returns
    -------
    data : list
        List of U64 numbers.

    """
    
    # Open .bin file
    with open(fname, mode='rb') as file:
        file.seek(start_index)
        stopIndex = start_index + length
        rawdata = list(file.read(stopIndex - start_index))

    # number of data points
    N = len(rawdata) / 8  # 8 bytes per U64 number

    # Create vector data with the U64 numbers
    data = []
    for i in range(round(N)):
        start = 8 * i
        newRawDataPoint = rawdata[start:start+8]
        newDataPoint = newRawDataPoint[0]
        for j in range(7):
            newDataPoint = (newDataPoint << 8) + newRawDataPoint[j+1]
        data.append(newDataPoint)

    # Return data vector
    return data



import numpy as np


def calcdist_from_coord(data, coord=None):
    """
    Calculate the distance from a fixed point to every cell of an array

    Parameters
    ----------
    data : np.array
        2d np.array (only the dimensions of the array are needed, the
        contents of the array can be whatever).
    coord : list, optional
        [idy, idx] indices of the data point from which to calculate
        the distance, default is center. The default is None.

    Returns
    -------
    dist : np.array
        2d np.array with for each cell the distance from that cell to
        the given coordinates.

    """

    dataShape = np.shape(data)
    
    Ny = dataShape[0]
    Nx = dataShape[1]

    if coord == None:
        # use center coordinate
        coord = [int(np.floor(Ny/2)), int(np.floor(Nx/2))]
    
    x = np.linspace(0, Nx-1, Nx)
    y = np.linspace(0, Ny-1, Ny)
    
    xv, yv = np.meshgrid(x, y)
    
    yv -= coord[0]
    xv -= coord[1]
    
    dist = np.sqrt(xv**2 + yv**2)

    return dist


def list_of_pixel_pairs_at_distance(dist, N=25, pixelsOff=[]):
    """
    Return list of pixel pairs located at a given shift vector from each other

    Parameters
    ----------
    dist : list
        two numbers indicating the shift in y and x.
    N : int, optional
        Total number of pixels in the square array. The default is 25.
    pixelsOff : list
        List of pixel numbers that are considered bad and should be removed from the list.

    Returns
    -------
    listPairs : list
        list of pixel pairs fullfilling the requirements.

    """
    listPairs = []
    for det1 in range(N):
        for det2 in range(N):
            if calc_shift_from_coord(det1, det2, int(np.sqrt(N))) == dist and det1 not in pixelsOff and det2 not in pixelsOff:
                listPairs.append([det1, det2])
    return listPairs


def calc_shift_from_coord(det1, det2, N=5):
    """
    Calculate shift vector between two detector elements

    Parameters
    ----------
    det1 : int
        pixel number.
    det2 : int
        second pixel number.
    N : int, optional
        Number of pixels in each direction of the array. The default is 5.

    Returns
    -------
    list
        two numbers with the y and x shift between the two pixels in arb. units.

    """
    # positive shifts are to the right and down
    row1 = np.floor(det1 / N)
    row2 = np.floor(det2 / N)
    col1 = np.mod(det1, N)
    col2 = np.mod(det2, N)
    
    deltay = int(row2 - row1)
    deltax = int(col2 - col1)
    
    return [deltay, deltax]